/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: af_hci.c,v 1.158 2003/03/03 11:42:22 kds Exp $

   AF_AFFIX - HCI Protocol Address family for socket interface

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/		

/* The following prevents "kernel_version" from being set in this file. */
#define __NO_VERSION__

#include <linux/config.h>
#include <linux/version.h>

/* Module related headers, non-module drivers should not include */
#include <linux/module.h>
#include <linux/init.h>

/* Standard driver includes */
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/timer.h>
#include <linux/spinlock.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/ioctl.h>
#include <linux/proc_fs.h>
#include <linux/notifier.h>
#include <linux/poll.h>
#include <linux/kmod.h>
#include <linux/file.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <linux/utsname.h>

#include <linux/skbuff.h>
#include <linux/socket.h>
#include <net/sock.h>

/* Local Includes */
#define FILEBIT	DBAFHCI

#include <affix/bluetooth.h>
#include <affix/btdebug.h>
#include <affix/hci.h>

/* include it for kernel HCI commands support */
static int	errno;
#include <affix/hci_cmds.h>


btlist_head_t	hci_socks;
spinlock_t	hpf_spinlock = SPIN_LOCK_UNLOCKED;

/*
 * debuging stuff
 */
static long 	sock_count = 0;

typedef struct  {
	struct list_head	q;

	struct sock	*sk;
	unsigned long	flags;
	hci_con		*con;
	hci_struct	*hci;
	__u64		event_mask;
	unsigned int	pkt_mask;

	/* command processing */
	hci_lock_t	*lock;		/* waiting here */
	__u16		opcode;

} hci_sock_t;

int hpf_cmd_handler(hci_sock_t *hsk, __u8 event, struct sk_buff *skb);

static inline hci_sock_t *hpf_get(struct sock *sk)
{
	return (hci_sock_t*)sk->prot;
}

static inline void hpf_unlock(hci_sock_t *hsk)
{
	if (hsk->lock) {
		hci_unlock(hsk->lock);
		hsk->lock = NULL;
	}
}

/***************************    Connection Management   ********************************/

hci_sock_t *hpf_alloc(int priority)
{
	hci_sock_t	*hsk;

	hsk = (hci_sock_t*)kmalloc(sizeof(hci_sock_t), priority);
	if (hsk == NULL)
		return NULL;
	memset(hsk, 0, sizeof(hci_sock_t));

	return hsk;
}

void hpf_destruct(struct sock *sk)
{
	hci_sock_t	*hsk = hpf_get(sk);

	DBFENTER;
	
	kfree(hsk);
	sock_count--;	/* XXX */

	DBFEXIT;
}

void hpf_destroy(struct sock *sk)
{
	hci_sock_t	*hsk = hpf_get(sk);
	
	DBFENTER;
	btl_unlink(&hci_socks, hsk);

	if (hsk->hci)
		hci_put(hsk->hci);
	
	if (hsk->con) {
		hcc_put(hsk->con);
		hsk->con = NULL;
	}
	if (sk->pair) {
		sock_put(sk->pair);
		sk->pair = NULL;
	}
	if (hsk->flags & AFFIX_FLAGS_CTLMASK)
		affix_ctl_mode &= ~(hsk->flags & AFFIX_FLAGS_CTLMASK);

	if (sk->protocol == BTPROTO_HCIACL || sk->protocol == BTPROTO_HCISCO) {
		struct sk_buff	*skb;

		DBPRT("queue len: %d\n", skb_queue_len(&sk->receive_queue));
		/* remove all packets and check if threre is a connection in the queue */
		while ((skb = skb_dequeue(&sk->receive_queue)) != NULL) {
			if (skb->sk != sk)	/* A pending connection */
				hpf_destroy(skb->sk);
			kfree_skb(skb);
		}
	}
	sock_put(sk);
	DBFEXIT;
}

struct sock *hpf_create(int family, int priority, int zero_it)
{
	struct sock	*sk;
	hci_sock_t	*hsk;

	DBFENTER;
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	sk  = sk_alloc(family, priority, zero_it);
#else
	sk  = sk_alloc(family, priority, zero_it, NULL);
#endif
	if (sk == NULL)
		return NULL;

	hsk = hpf_alloc(priority);
	if (hsk == NULL) {
		sk_free(sk);
		return NULL;
	}

	hsk->sk = sk;
	sk->prot = (void*)hsk;

	//__btl_add_tail(&hci_socks, hsk);

	sock_count++;	/* XXX */

	DBFEXIT;
	return sk;
}

void hpf_init_data(struct socket *sock, struct sock *sk)
{
	DBFENTER;
	
	sock_init_data(sock, sk);
	sk->destruct = hpf_destruct;
	
	DBFEXIT;
}

int hpf_wait_for_state(struct sock *sk, __u8 state, int noblock)
{
	int	err = 0;
	long	timeo;
	DECLARE_WAITQUEUE(wait, current);

	DBFENTER;

	timeo = sock_sndtimeo(sk, noblock);

	add_wait_queue(sk->sleep, &wait);
	set_current_state(TASK_INTERRUPTIBLE);
	while (sk->state != state) {
		if (!timeo) {
			err = -EAGAIN;
			break;
		}
		if ((err = sock_error(sk))) 
			break;
		if (signal_pending(current)) {
			err = -ERESTARTSYS;
			break;
		}
		timeo = schedule_timeout(timeo);
		set_current_state(TASK_INTERRUPTIBLE);
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(sk->sleep, &wait);

	DBFEXIT;
	return err;
}

int hpf_connect_req(struct sock *sk, struct sockaddr_affix *addr)
{
	int		err = 0;
	hci_struct	*hci;
	hci_sock_t	*hsk = hpf_get(sk);

	DBFENTER;

	if (addr->devnum) {
		hci = hci_lookup_devnum(addr->devnum);	/* find device */
		if (hci == NULL)
			return -ENODEV;
	} else
		hci = NULL;

	if (sk->protocol == BTPROTO_HCIACL) {
		err = lp_connect_req(hci, &addr->bda, &hsk->con);
		if (err) {
			if (hci)
				hci_put(hci);
			return err;
		}
		hsk->hci = hci;
		if (hcc_linkup(hsk->con))
			sk->state = CON_ESTABLISHED;

	} else if (sk->protocol == BTPROTO_HCISCO) {
		err = lp_add_sco(hci, &addr->bda, &hsk->con);
		if (err) {
			if (hci)
				hci_put(hci);
			return err;
		}
		hsk->hci = hci;
	}
	DBFEXIT;
	return err;
}

void hpf_shutdown(struct sock *sk, int reason)
{
	if (reason < 0 )
		sk->err = -reason;
	else if (reason == 0xEEEE)
		sk->err = ETIMEDOUT;
	else 
		sk->err = ECONNRESET;

        sk->shutdown |= SEND_SHUTDOWN;
	sk->state = CON_CLOSED;
	if (!sk->dead) {
                sk->dead = 1;
		sk->state_change(sk);
        }
}

/* 
   HPF protocol subsystem.  Client of HCI. Callback functions
*/
int hpf_recv_raw(hci_struct *hci, hci_con *con, struct sk_buff *skb)
{
	hci_sock_t		*hsk, *next;
	struct sk_buff		*new_skb;
	struct sock		*sk;

	DBFENTER;

#ifdef CONFIG_AFFIX_L2CAP
	if (con && con->link_type == HCI_LT_ACL) {
		/* ACL L2CAP packet */
		struct sk_buff		*acl;
		hci_acl_hdr_t		*hdr;
		
		acl = skb_clone(skb, GFP_ATOMIC);
		if (acl) {
			hdr = (void*)acl->data;
			skb_pull(acl, HCI_ACL_HDR_LEN);
			lp_receive_data(con, HCI_PB_FLAG(hdr->Connection_Handle) == HCI_PB_FIRST,
					HCI_BC_FLAG(hdr->Connection_Handle), acl);
		}
	}
#endif
	btl_read_lock(&hci_socks);
	btl_for_each_safe (hsk, hci_socks, next) {
		sk = hsk->sk;
		if (sk->type != SOCK_RAW) {
			/* HCIACL or HCISCO sockets - connection oriented */
			if (!con || hsk->con != con)	/* this check is enough?? */
				continue;
		} else {
			if (hci_skb(skb)->outgoing && !(hsk->flags & AFFIX_FLAGS_PROMISC))
				continue;
				
			if (!test_bit(skb->pkt_type, &hsk->pkt_mask))
				continue;
				
			if (hsk->hci && hsk->hci != hci)
				continue;
			
			if (skb->pkt_type == HCI_EVENT) {
				__u8	event = *skb->data;
				if (!test_bit64((event - 1) & 0x3F, &hsk->event_mask))
					continue;
					
				if (hsk->flags & AFFIX_FLAGS_W4_STATUS) {
					/* command executing socket */
					if (hpf_cmd_handler(hsk, event, skb) != 0) /* not ower command */
						continue;
				}
			}
		}
		if (&hci_socks != (void*)next) {
			new_skb = skb_clone(skb, GFP_ATOMIC);
			if (new_skb == NULL)
				break;
		} else {
			/* last element - do not need to clone */
			new_skb = skb_get(skb);
		}
		affix_sock_rx(hsk->sk, new_skb);
	}
	btl_read_unlock(&hci_socks);
	kfree_skb(skb);
	DBFEXIT;
	return 0;
}

int hci_deliver_msg(void *msg, int size)
{
	struct sk_buff	*skb;
	
	skb = alloc_skb(1 + size, gfp_any());
	if (skb == NULL)
		return -ENOMEM;
	skb_reserve(skb, 1);
	memcpy(skb_put(skb, size), msg, size);
	skb->pkt_type = HCI_MGR;
	return hpf_recv_raw(NULL, NULL, skb);
}

struct sock *hpf_dup(struct sock *osk)
{
	struct sock	*sk;

	sk = hpf_create(PF_AFFIX, GFP_ATOMIC, 1);
	if (sk == NULL)
		return NULL;

	hpf_init_data(NULL, sk);

	sk->pair = osk;

	sk->protocol = osk->protocol;
	sk->type     = osk->type;
	sk->destruct = osk->destruct;
	sk->sleep    = osk->sleep;
	sk->priority = osk->priority;
	//	sk->socket   = osk->socket;

	return sk;
}

int hpf_connect_cfm(hci_con *con, int status)
{
	int		err = 0;
	struct sock	*sk, *pair;
	hci_sock_t	*hsk, *phsk;
	struct sk_buff	*skb;
	int		type;
	
	DBFENTER;
#ifdef CONFIG_AFFIX_L2CAP
	if (con->link_type == HCI_LT_ACL)
		lp_connect_cfm(con, status);
#endif
	btl_write_lock(&hci_socks);
	btl_for_each (phsk, hci_socks) {
		pair = phsk->sk;
		if (pair->socket == NULL)
			continue;
		if (pair->dead)
			continue;
		if (pair->protocol == BTPROTO_HCIACL)
			type = HCI_LT_ACL;
		else if (pair->protocol == BTPROTO_HCISCO )
			type = HCI_LT_SCO;
		else
			continue;
		if (con->link_type != type)
			continue;
		if (pair->state == CON_LISTEN) {
			if ((pair->bound_dev_if == 0 || pair->bound_dev_if == con->hci->devnum)) {
				if (pair->ack_backlog < pair->max_ack_backlog && status == 0) {
					DBPRT("Server socket found\n");
					sk = hpf_dup(pair);
					if (sk == NULL) {
						err = -ENOMEM;
						break;
					}
					__btl_add_tail(&hci_socks, hpf_get(sk));
					sock_hold(pair);
				} else
					continue;
			} else
				continue;
		} else if (pair->state == CON_CONNECTING) {
			if (phsk->con == con) {
				sk = pair;
				if (status) {
					hpf_shutdown(sk, status);
					continue;
				}
			} else
				continue;
		} else
			continue;

		hsk = hpf_get(sk);
		// set attributes for hci socket
		sk->bound_dev_if = con->hci->devnum;
		sk->state = CON_ESTABLISHED;
		if (hsk->hci == NULL) {
			hsk->hci = con->hci;
			hci_hold(con->hci);
		}
		if (hsk->con == NULL) {
			hsk->con = con;
			hcc_hold(con);
		}
		if (sk->pair){ /* incomming connection */
			sk->pair->ack_backlog++;
			skb = alloc_skb(0, GFP_ATOMIC); 
			skb_set_owner_r(skb, sk);
			skb_queue_tail(&sk->pair->receive_queue, skb);
			sk->pair->data_ready(sk->pair, skb->len);
		} else
			sk->state_change(sk);
	}
	btl_write_unlock(&hci_socks);
	DBFEXIT;
	return 0;
}

int hpf_disconnect_ind(hci_con *con)
{
	hci_sock_t	*hsk;

	DBFENTER;
#ifdef CONFIG_AFFIX_L2CAP	
	lp_disconnect_ind(con);
#endif
	btl_read_lock(&hci_socks);
	btl_for_each (hsk, hci_socks) {
		if (hsk->con == con) {
			if (!hsk->sk->dead)
				hpf_shutdown(hsk->sk, 0);
		}
	}
	btl_read_unlock(&hci_socks);
	DBFEXIT;
	return 0;
}


/******************************  HCI PROTOCOL FAMILY SUBSYSTEM  **************************/

static inline void hpf_wait_on_queue(struct sock *sk)
{
	DECLARE_WAITQUEUE(wait, current);

	DBFENTER;

	add_wait_queue(sk->sleep, &wait);
	set_current_state(TASK_INTERRUPTIBLE);

	if (skb_peek(&sk->receive_queue) == NULL)
		schedule();

	set_current_state(TASK_RUNNING);
	remove_wait_queue(sk->sleep, &wait);

	DBFEXIT;
}

int hpf_release(struct socket *sock)
{
	struct sock	*sk = sock->sk;
	hci_sock_t	*hsk;
	int		dec = 1;

	DBFENTER;

	if (!sk)
		return 0;
	
	sock_orphan(sk);	/* important */
        sock->sk = NULL;

	hsk = hpf_get(sk);
	if (hsk->flags & AFFIX_FLAGS_KERNEL)
		dec = 0;
	if (hsk->flags & AFFIX_FLAGS_SUPER) {
		if (hsk->hci) {
			hci_struct	*hci = hsk->hci;
			hci->open_count--;
			if (!hci->open_count && !test_bit(HCIDEV_STATE_START, &hci->state))
				hcidev_close(hci);
		}
	}
	sk->state = CON_CLOSED;
        sk->err = 0;
	sk->shutdown |= SEND_SHUTDOWN;
	hpf_destroy(sk);
	if (dec)
		MOD_DEC_USE_COUNT;
	DBFEXIT;
	return 0;
}

int hpf_bind(struct socket *sock, struct sockaddr *addr, int addr_len)
{
	int		err = 0;
	struct sockaddr_affix	*saddr = (void*)addr;

	DBFENTER;
	sock->sk->bound_dev_if = saddr->devnum;
	DBFEXIT;
	return err;
}

/*
    Connect to remote device
    Wait until connection established
*/
int hpf_connect(struct socket *sock, struct sockaddr *addr, int alen, int flags)
{
	struct sock		*sk = sock->sk;
	struct sockaddr_affix	*saddr = (void*)addr;
	int			err = 0;

	DBFENTER;

	/* deal with restarts */
	if (sk->state == CON_ESTABLISHED && sock->state == SS_CONNECTING) {
		sock->state = SS_CONNECTED;
		return 0;
	}
		
	if (sk->state == CON_CLOSED && sock->state == SS_CONNECTING) {
		sock->state = SS_UNCONNECTED;
		return -ECONNREFUSED;
	}
	
	if (sk->state == CON_ESTABLISHED)
		return -EISCONN;      /* No reconnect on a seqpacket socket */
	
	/* Move to connecting socket, start sending Connect Requests */
	sock->state = SS_CONNECTING;
	sk->state   = CON_CONNECTING;

	err = hpf_connect_req(sk, saddr);
	if (err < 0)
		goto exit;

	err = hpf_wait_for_state(sk, CON_ESTABLISHED, flags & O_NONBLOCK);
	if (err)
		goto exit;

	if (sk->state == CON_ESTABLISHED) {
		sock->state = SS_CONNECTED;
	}

 exit:
	DBFEXIT;
	return err;
}

int hpf_open_hci(struct sock *sk, struct hci_open *param)
{
	int			err = 0;
	hci_struct		*hci = NULL;
	hci_sock_t		*hsk = hpf_get(sk);

	DBFENTER;

	if (param->cmd & HCI_OPEN_KERNEL) {
		hsk->flags |= AFFIX_FLAGS_KERNEL;
		MOD_DEC_USE_COUNT;
	}
	if (param->cmd & HCI_OPEN_SUPER)
		hsk->flags |= AFFIX_FLAGS_SUPER;

	hsk->flags &= ~AFFIX_FLAGS_TYPE;

	switch (param->cmd & ~HCI_OPEN_MASK) {
	case HCI_OPEN_NAME:
		btl_read_lock(&hcidevs);
		btl_for_each (hci, hcidevs) {
			if (strncmp(hci->name, param->name, IFNAMSIZ) == 0 &&
				(hci_running(hci) || (param->cmd & HCI_OPEN_SUPER))) {
				hci_hold(hci);
				break;
			}
		}
		btl_read_unlock(&hcidevs);
		if (!hci)
			return -ENODEV;
		break;
	case HCI_OPEN_ID:
		btl_read_lock(&hcidevs);
		btl_for_each (hci, hcidevs) {
			if (hci->devnum == param->devnum &&
				(hci_running(hci) || (param->cmd & HCI_OPEN_SUPER))) {
				hci_hold(hci);
				break;
			}
		}
		btl_read_unlock(&hcidevs);
		if (!hci)
			return -ENODEV;
		break;
	case HCI_OPEN_EVENT:
		hsk->pkt_mask |= (1 << HCI_EVENT);
		break;
	case HCI_OPEN_MGR:
		hsk->pkt_mask |= (1 << HCI_MGR);
		break;
	}
	if (hci && (hsk->flags & AFFIX_FLAGS_SUPER)) {
		/* open net device */
		err = hcidev_open(hci);
		if (err) {
			hci_put(hci);
			return err;
		}
		hci->open_count++;
	}
	hsk->hci = hci;
	DBFEXIT;
	return err;
}

int hpf_listen_event(struct sock *sk, __u64 *event_mask)
{
	int		err = 0;
	hci_sock_t	*hsk = hpf_get(sk);

	DBFENTER;
	hsk->event_mask = *event_mask;
	DBFEXIT;
	return err;
}


/* ------------------- device specific stuff ------------------ */

static inline int hpf_set_audio(hci_struct *hci, struct affix_audio *audio)
{
	int 	err = 0;
	int	fd;

	DBFENTER;

	if (!hci)
		return -ENODEV;

	hci->audio_mode = audio->mode;
	if (hci->audio_mode & (AFFIX_AUDIO_IN | AFFIX_AUDIO_OUT)) {
		fd = hci_open_id(hci->devnum);
		if (fd < 0)
			return fd;
		if (hci->audio_mode & AFFIX_AUDIO_FLOW) {
			/* enable SCO flow control */
			err = HCI_WriteSCOFlowControlEnable(fd, 0x01);
			if (err)
				hci->audio_mode &= ~AFFIX_AUDIO_FLOW;
		} else
			/* disable SCO flow control */
			err = HCI_WriteSCOFlowControlEnable(fd, 0x00);
		if (err >= 0 && audio->setting) {
			hci->audio_setting = audio->setting;
			err = HCI_WriteVoiceSetting(fd, hci->audio_setting);
		}
		err = (err<0)?-errno:err;
		hci_close(fd);
		if (err < 0)
			return err;
	}
	err = hcidev_ioctl(hci, SIOCHCI_SET_AUDIO, NULL);
	DBFEXIT;
	return err;
}

static inline int hpf_hci_disc(hci_struct *hci, struct sockaddr_affix *sa)
{
	hci_con		*con;

	DBFENTER;
	con = hcc_lookup_acl(hci, &sa->bda);
	if (!con)
		return -EINVAL;
	lp_disconnect_req(con, 0x13);
	hcc_put(con);
	__affix_yield();	// allow affixd to get an event
	DBFEXIT;
	return 0;
}

static inline int hpf_get_conn(hci_struct *hci, struct affix_conn_info *ci)
{
	hci_con		*con;

	DBFENTER;
	if (!hci)
		return -ENODEV;
	con = hcc_lookup_acl(hci, &ci->bda);
	if (!con)
		return -EINVAL;
	ci->dport = con->chandle;
	hcc_put(con);
	DBFEXIT;
	return 0;
}

int hpf_get_devs(int *devs)
{
	hci_struct	*hci;
	int		count = 0;

	DBFENTER;
	btl_read_lock(&hcidevs);
	btl_for_each (hci, hcidevs) {
		if (hci->deadbeaf)
			continue;
		devs[count++] = hci->devnum;
		if (count == HCI_MAX_DEVS)
			break;
	}
	btl_read_unlock(&hcidevs);
	DBFEXIT;
	return count;
}

int hpf_set_secmode(hci_struct *hci, int sec_mode);

static inline int hpf_setget_attr(hci_struct *hci, int get, struct hci_dev_attr *da)
{
	DBFENTER;
	if (!hci) {
		da->name[IFNAMSIZ-1] = '\0';
		btl_read_lock(&hcidevs);
		btl_for_each (hci, hcidevs) {
			if (((da->devnum && da->devnum == hci->devnum) || (!da->devnum && strncmp(hci->name, da->name, IFNAMSIZ) == 0))
						 && !hci->deadbeaf) {
				hci_hold(hci);
				break;
			}
		}
		btl_read_unlock(&hcidevs);
		if (!hci)
			return -ENODEV;
	} else
		hci_hold(hci);
	if (get) {
		snprintf(da->name, IFNAMSIZ, "%s", hci->name);
		da->bda = hci->bda;
		da->devnum = hci->devnum;
		da->flags = hci->flags;
		da->pkt_type = hci->pkt_type;
		da->stats = hci->stats;
	}
	hci_put(hci);
	DBFEXIT;
	return 0;
}

static inline int hpf_setget_ctl(struct sock *sk, int get, int *mode)
{
	int 		err = 0;
	hci_sock_t	*hsk = hpf_get(sk);

	DBFENTER;
	if (get) {
		*mode = affix_ctl_mode;
		return 0;
	}
	if (affix_ctl_mode & *mode)
		return -EBUSY;
	affix_ctl_mode |= *mode;
	hsk->flags |= *mode;
	DBFEXIT;
	return err;

}

int hpf_start_dev(hci_struct *hci)
{
	int	err, fd, timeout;
	__u8	hci_version, lmp_version;
	__u16	hci_revision, manfid, lmp_subversion;

	DBFENTER;

	if (test_bit(HCIDEV_STATE_START, &hci->state))
		return 0;
			
	hci->pkt_type = HCI_PT_DM1 | HCI_PT_DH1;

	fd = _hci_open_id(hci->devnum);
	if (fd < 0)
		return -ENODEV;

	DBPRT("Device opened...\n");
	//set_current_state(TASK_INTERRUPTIBLE);
	//schedule_timeout(2*HZ);
	
	/* try first to wait a bit */
	timeout = sysctl_hci_req_timeout;
	sysctl_hci_req_timeout = 10 * HZ;
	err = HCI_ReadBDAddr(fd, &hci->bda);
	if (err)
		goto initerr;
	sysctl_hci_req_timeout = timeout;

	err = HCI_ReadBufferSize(fd, &hci->acl_mtu, &hci->sco_mtu, &hci->acl_num, &hci->sco_num);
	if (err)
		goto initerr;

	err = HCI_ReadLocalVersionInformation(fd, &hci_version, &hci_revision, 
						&lmp_version, &manfid, &lmp_subversion);
	if (err)
		goto initerr;

	hci->manfid = manfid;

	err = HCI_ReadLocalSupportedFeatures(fd, &hci->lmp_features);
	if (err)
		goto initerr;

	/* some general settings */
	err = HCI_WriteClassOfDevice(fd, HCI_COD_INFORMATION | HCI_COD_NETWORKING | HCI_COD_COMPUTER | HCI_COD_HANDPC);
	if (err)
		goto initerr;
		
	/* set some parameters */
	down_read(&uts_sem);
	err = HCI_ChangeLocalName(fd, system_utsname.nodename);
	up_read(&uts_sem);
	if (err)
		goto initerr;

	err = hpf_set_secmode(hci, HCI_SECURITY_SERVICE);
	if (err)
		goto initerr;

	err = HCI_WriteScanEnable(fd, hci->flags>>HCI_FLAGS_SCAN_BITS);	// not visible by default
	if (err)
		goto initerr;
		
	if (hci->lmp_features & HCI_LF_3SLOTS)
		hci->pkt_type |= HCI_PT_DM3 | HCI_PT_DH3;
	if (hci->lmp_features & HCI_LF_5SLOTS)
		hci->pkt_type |= HCI_PT_DM5 | HCI_PT_DH5;
	if (hci->lmp_features & HCI_LF_SCO) {
		hci->pkt_type |= HCI_PT_HV1;
		if (hci->lmp_features & HCI_LF_HV2)
			hci->pkt_type |= HCI_PT_HV2;
		if (hci->lmp_features & HCI_LF_HV3)
			hci->pkt_type |= HCI_PT_HV3;
	}
	if (!(hci->lmp_features & HCI_LF_SWITCH)) {
		hci->flags |= HCI_ROLE_DENY_SWITCH;	// deny switch if no support
		hci->flags &= ~HCI_ROLE_BECOME_MASTER;
	}
#ifdef CONFIG_AFFIX_DTL1_FIX
	if (manfid == HCI_LMP_NOKIA) {
		hci->fixit = 1;
		/*  Removed when UART RI flow control was introduced ??? */
		hci->acl_num = 1;
	}
#endif
	atomic_set(&hci->acl_count, hci->acl_num);
	atomic_set(&hci->sco_count, hci->sco_num);

	hci_close(fd);
	set_bit(HCIDEV_STATE_START, &hci->state);
	hci->flags |= HCI_FLAGS_UP;
	hcidev_state_change(hci, HCIDEV_UP);
	DBFEXIT;
	return 0;
initerr:
	err = (err<0)?-errno:err;
	hci_close(fd);
	BTERROR("Bluetooth device initialization failed: %d\n", err);
	return err;
}

int hpf_stop_dev(hci_struct *hci)
{
	int	err, fd;
	
	DBFENTER;

	if (!test_and_clear_bit(HCIDEV_STATE_START, &hci->state))
		return 0;
			
	hci->flags &= ~HCI_FLAGS_UP;
	hci->flags = 0;	//XXX:

	fd = _hci_open_id(hci->devnum);
	if (fd < 0)
		return -ENODEV;
	err = HCI_Reset(fd);	// reset
	hci_close(fd);

	hcidev_state_change(hci, HCIDEV_DOWN);
	DBFEXIT;
	return err;
}

	
static inline int hpf_setget_flags(hci_struct *hci, int get, int *flags)
{
	int 	err = 0;

	DBFENTER;

	if (!hci)
		return -ENODEV;
	if (get) {
		*flags = hci->flags;
		return 0;
	}
	/* up/down */
	if (!(hcidev_running(hci)))
		return -EINVAL;
	if (*flags & HCI_FLAGS_UP)
		err = hpf_start_dev(hci);
	else
		err = hpf_stop_dev(hci);
	DBFEXIT;
	return err;
}

static inline int hpf_setget_pkttype(hci_struct *hci, int get, int *pkt_type)
{
	DBFENTER;
	if (!hci)
		return -ENODEV;
	if (get) {
		*pkt_type = hci->pkt_type;
		return 0;
	}
	hci->pkt_type = (*pkt_type) ? *pkt_type:(HCI_PT_DM1|HCI_PT_DH1|HCI_PT_DM3|HCI_PT_DH3|HCI_PT_DM5|HCI_PT_DH5|HCI_PT_HV1);
	DBFEXIT;
	return 0;
}

int hpf_set_secmode(hci_struct *hci, int sec_mode)
{
	int	err = 0, fd;

	DBFENTER;
	if (!hci)
		return -ENODEV;
	
	sec_mode &= HCI_FLAGS_SECURITY;
	if (!sec_mode)
		return -EINVAL;
		
	if ((hci->flags & HCI_FLAGS_SECURITY) ^ sec_mode) {
		hci->flags = (hci->flags & ~HCI_FLAGS_SECURITY) | sec_mode;
		
		if (!hci_running(hci))
			return 0;
		fd = hci_open_id(hci->devnum);
		if (fd < 0)
			return -ENODEV;
		err = HCI_WriteAuthenticationEnable(fd, (hci->flags & HCI_SECURITY_AUTH) != 0);
		if (err) {
			err = (err<0)?-errno:err;
			hci_close(fd);
			return err;
		}
		err = HCI_WriteEncryptionMode(fd, (hci->flags & HCI_SECURITY_ENCRYPT) != 0);
		err = (err<0)?-errno:err;
		hci_close(fd);
	}
	DBFEXIT;
	return err;
}

static inline int hpf_set_role(hci_struct *hci, int role)
{
	DBFENTER;
	if (!hci)
		return -ENODEV;
	role &= HCI_FLAGS_ROLE;
	hci->flags = (hci->flags & ~HCI_FLAGS_ROLE) | role;
	if (!(hci->lmp_features & HCI_LF_SWITCH)) {
		hci->flags |= HCI_ROLE_DENY_SWITCH;	// deny switch if no support
		hci->flags &= ~HCI_ROLE_BECOME_MASTER;
	}
	DBFEXIT;
	return 0;
}

int hpf_set_scan(hci_struct *hci, int scan_mode)
{
	int	err = 0, fd;

	DBFENTER;
	if (!hci)
		return -ENODEV;
	
	scan_mode &= HCI_FLAGS_SCAN;
		
	if ((hci->flags & HCI_FLAGS_SCAN) ^ scan_mode) {
		hci->flags = (hci->flags & ~HCI_FLAGS_SCAN) | scan_mode;
		if (!hci_running(hci))
			return 0;
		fd = hci_open_id(hci->devnum);
		if (fd < 0)
			return -ENODEV;
		err = HCI_WriteScanEnable(fd, scan_mode>>HCI_FLAGS_SCAN_BITS);
		err = (err<0)?-errno:err;
		hci_close(fd);
	}
	DBFEXIT;
	return err;
}

static inline int hpf_setup_uart(hci_struct *hci, struct open_uart *uart)
{
	int	err;

	DBFENTER;
	btl_read_lock(&hcidevs);
	btl_for_each (hci, hcidevs) {
		if (strncmp(hci->name, uart->dev, IFNAMSIZ) == 0 
				&& hcidev_present(hci)) {
			hci_hold(hci);
			break;
		}
	}
	btl_read_unlock(&hcidevs);
	if (!hci)
		return -ENODEV;
	err = hcidev_ioctl(hci, BTIOC_SETUP_UART, uart);
	hci_put(hci);
	DBFEXIT;
	return err;
}

static struct affix_pan_operations	affix_pan_ops;

int affix_set_pan(struct affix_pan_operations *ops)
{
	DBFENTER;
	
	if (ops == NULL)
		affix_pan_ops.owner = NULL;
	else
		affix_pan_ops = *ops;
	
	DBFEXIT;
	return 0;
}


int hpf_init_pan(struct pan_init *arg)
{
	int	err = -ENODEV;

	DBFENTER;
#if defined(CONFIG_AFFIX_PAN_MODULE)
	if (!affix_pan_ops.owner) {// || !try_inc_mod_count(affix_pan_ops.owner)) { 
		request_module("affix_pan");
		if (!affix_pan_ops.owner)// || !try_inc_mod_count(affix_pan_ops.owner))
			return -ENODEV;
	}
#elif !defined(CONFIG_AFFIX_PAN)
		return -EPROTONOSUPPORT;
#endif
	if (affix_pan_ops.ioctl)
		err = affix_pan_ops.ioctl(BTIOC_PAN_INIT, (unsigned long)arg);
	//if (affix_uart_ops.owner)
	//	__MOD_DEC_USE_COUNT(affix_uart_ops.owner);
	DBFEXIT;
	return err;
}


/*
    our system calls for user space programs
*/
int hpf_ioctl(struct socket *sock, unsigned int cmd,  unsigned long arg)
{
	struct sock	*sk = sock->sk;
	hci_sock_t	*hsk = hpf_get(sk);
	int		err = 0;

	DBFENTER;

	if (_IOC_TYPE(cmd) == BTIOC_MAGIC) {
		switch(cmd) {
			case BTIOC_OPEN_HCI:
				err = hpf_open_hci(sk, (struct hci_open*)arg);
				break;
			case BTIOC_DBMGET:
				*(__u32*)arg = affix_dbmask;
				break;
			case BTIOC_DBMSET:
				affix_dbmask = *(__u32*)arg;
				break;
			case BTIOC_ADDPINCODE:
				err = affix_add_pin((struct PIN_Code *)arg);
				break;
			case BTIOC_REMOVEPINCODE:
				err = affix_remove_pin((BD_ADDR*)arg);
				break;
			case BTIOC_ADDLINKKEY:
				err = affix_add_key((struct link_key*)arg);
				break;
			case BTIOC_REMOVELINKKEY:
				err = affix_remove_key((BD_ADDR*)arg);
				break;
			case BTIOC_SETUP_UART:
				err = hpf_setup_uart(hsk->hci, (struct open_uart*)arg);
				break;
			case BTIOC_OPEN_UART:
				err = affix_open_uart((struct open_uart*)arg);
				break;
			case BTIOC_CLOSE_UART:
				err = affix_close_uart((struct open_uart*)arg);
				break;
			case BTIOC_GETDEVS:
				err = hpf_get_devs((int*)arg);
				break;
			case BTIOC_SET_CTL:
				err = hpf_setget_ctl(sk, 0, (int*)arg);
				break;
			case BTIOC_GET_CTL:
				err = hpf_setget_ctl(sk, 1, (int*)arg);
				break;
			case BTIOC_SET_AUDIO:
				err = hpf_set_audio(hsk->hci, (struct affix_audio*)arg);
				break;
			case BTIOC_HCI_DISC:
				err = hpf_hci_disc(hsk->hci, (struct sockaddr_affix*)arg);
				break;
			case BTIOC_GET_CONN:
				err = hpf_get_conn(hsk->hci, (struct affix_conn_info*)arg);
				break;
			case BTIOC_GET_ATTR:
				err = hpf_setget_attr(hsk->hci, 1, (struct hci_dev_attr*)arg);
				break;
			case BTIOC_SET_ATTR:
				err = hpf_setget_attr(hsk->hci, 0, (struct hci_dev_attr*)arg);
				break;
			case BTIOC_SET_FLAGS:
				err = hpf_setget_flags(hsk->hci, 0, (int*)arg);
				break;
			case BTIOC_GET_FLAGS:
				err = hpf_setget_flags(hsk->hci, 1, (int*)arg);
				break;
			case BTIOC_SET_PKTTYPE:
				err = hpf_setget_pkttype(hsk->hci, 0, (int*)arg);
				break;
			case BTIOC_GET_PKTTYPE:
				err = hpf_setget_pkttype(hsk->hci, 1, (int*)arg);
				break;
			case BTIOC_SET_SECMODE:
				err = hpf_set_secmode(hsk->hci, *(int*)arg);
				break;
			case BTIOC_SET_ROLE:
				err = hpf_set_role(hsk->hci, *(int*)arg);
				break;
			case BTIOC_SET_SCAN:
				err = hpf_set_scan(hsk->hci, *(int*)arg);
				break;
				/*
				 * PAN stuff
				 */
			case BTIOC_PAN_INIT:
				err = hpf_init_pan((struct pan_init*)arg);
				break;
			case BTIOC_PAN_CONNECT:
				if (affix_pan_ops.ioctl)
					err = affix_pan_ops.ioctl(cmd, (unsigned long)arg);
				else
					err = -ENODEV;
				break;
			default:
				return -ENOIOCTLCMD;
		}
	} else {
		/* deliver it to device driver */
		if (hsk->hci)
			err = hcidev_ioctl(hsk->hci, cmd, (void*)arg);
	}
	DBFEXIT;
	return err;
}


/*
   register protocol and allow clients to establish connection
*/
int hpf_listen(struct socket *sock, int backlog)
{
	struct sock	*sk = sock->sk;

	DBFENTER;

	if (sock->state != SS_UNCONNECTED)
		return -EINVAL;

	if (sk->state != CON_LISTEN) {
		sk->max_ack_backlog = backlog;
		sk->state = CON_LISTEN;
		sock->flags |= __SO_ACCEPTCON;
	}
	
	DBFEXIT;
	return 0;
}


/*
   Accept connection from remote device
   Wait until ready connection will become available
*/

int hpf_accept(struct socket *sock, struct socket *newsock, int flags)
{
	struct sock	*sk = sock->sk, *newsk;
	struct sk_buff	*skb;
	int		err = -EINVAL;

	DBFENTER;

	if (sock->state != SS_UNCONNECTED || !(sock->flags & __SO_ACCEPTCON))
		goto exit;
	
	while( (skb = skb_dequeue(&sk->receive_queue)) == NULL) {
		if (flags & O_NONBLOCK)
			return -EWOULDBLOCK;

		hpf_wait_on_queue(sk);
		if (signal_pending(current)) 
			return -ERESTARTSYS;
	}

	if (newsock->sk) {
		sk_free(newsock->sk);
		newsock->sk = NULL;
	}

	DBPRT("queue len: %d\n", skb_queue_len(&sk->receive_queue));
	
	newsk = skb->sk;
	sock_put(newsk->pair);		/* put pair */
	newsk->pair = NULL;
	sock_graft(newsk, newsock);
	newsock->state = SS_CONNECTED;
	sk->ack_backlog--;
	kfree_skb(skb);
	MOD_INC_USE_COUNT;
	err = 0;
 exit:			
	DBFEXIT;
	return err;
}

int hpf_getname(struct socket *sock, struct sockaddr *uaddr, int *alen, int peer)
{
	struct sock		*sk = sock->sk;
	struct sockaddr_affix	*addr = (struct sockaddr_affix*)uaddr;
	hci_sock_t		*hsk = hpf_get(sk);

	DBFENTER;
	addr->family = AF_AFFIX;
	addr->devnum = sk->bound_dev_if;
	if (hsk->con) {
		addr->bda = hsk->con->bda;
		addr->port = hsk->con->chandle;
	}
	*alen = sizeof(struct sockaddr_affix);
	DBFEXIT;
	return 0;

}

int hpf_setsockopt(struct socket *sock, int level, int optname, char *optval, int optlen)
{
	struct sock	*sk = sock->sk;
	hci_sock_t	*hsk = hpf_get(sk);

	DBFENTER;

	//lock_sock(sk);
	
	if (level != SOL_AFFIX)
		return -EOPNOTSUPP;
	switch (optname) {
		case BTSO_EVENT_MASK:
			if (copy_from_user(&hsk->event_mask, optval, sizeof(hsk->event_mask)))
				return -EFAULT;
			break;
		case BTSO_PKT_MASK:
			if (copy_from_user(&hsk->pkt_mask, optval, sizeof(hsk->pkt_mask)))
				return -EFAULT;
			break;
		default:
			return -ENOPROTOOPT;
	}
	//release_sock(sk);
	DBFEXIT;
	return 0;
}

int hpf_cmd_handler(hci_sock_t *hsk, __u8 event, struct sk_buff *skb)
{
	struct Command_Complete_Event	*cce;
	struct Command_Status_Event	*cse;
	struct sock			*sk = hsk->sk;

	DBFENTER;

	if (sk->err) {
		hsk->event_mask = 0;
		hpf_unlock(hsk);
		return -1;
	}
	cce = (void*)skb->data;
	cse = (void*)skb->data;

	if (event == HCI_E_COMMAND_COMPLETE) {
		if (cce->Command_Opcode != hsk->opcode)
			return -1;
		hsk->event_mask &= ~COMMAND_COMPLETE_MASK;/* we need only one */
		hpf_unlock(hsk);
	} else if (event == HCI_E_COMMAND_STATUS) {
		if (cse->Command_Opcode != hsk->opcode)
			return -1;
		hsk->event_mask &= ~COMMAND_STATUS_MASK;/* we need only one */
		if (cse->Status != 0)
			hsk->event_mask = 0;
		hpf_unlock(hsk);
	}
	DBFEXIT;
	return 0;// accept event
}

int hpf_wait_for_completion(struct sock *sk)
{
	int		err = 0;
	struct sk_buff	*skb;
	hci_sock_t	*hsk = hpf_get(sk);
	long		timeo = sk->rcvtimeo;	//old timeout

	DBFENTER;
	sk->rcvtimeo = sysctl_hci_req_timeout;	// default 5 sec.
	while ((hsk->flags & AFFIX_FLAGS_W4_STATUS) && err == 0) {
		skb = skb_recv_datagram(sk, 0, 0, &err);
		if (skb == NULL) {
			if (err != -ERESTARTSYS) //sock_intr_errno(timeo)
				hsk->flags &= ~AFFIX_FLAGS_CMD_PENDING;
			if (err == -EAGAIN)
				err = -EIO;
			if (err)
				break;
		} else {
			struct Command_Status_Event	*cse = (void*)skb->data;
			
			if (cse->hdr.EventCode == HCI_E_COMMAND_COMPLETE) {
				hsk->flags &= ~AFFIX_FLAGS_W4_STATUS;
				/* results of the command */
				skb_queue_head(&sk->receive_queue, skb);
				continue;
			} else if (cse->hdr.EventCode == HCI_E_COMMAND_STATUS) {
				hsk->flags &= ~AFFIX_FLAGS_W4_STATUS;
				err = cse->Status;
			}
			kfree_skb(skb);
		}
	}
	sk->rcvtimeo = timeo;
	DBFEXIT;
	return err;
}

int hpf_send_request(struct sock *sk, struct sk_buff *skb, int flags)
{
	hci_sock_t	*hsk = hpf_get(sk);
	int		err;

	if (hsk->flags & AFFIX_FLAGS_CMD_PENDING)
		goto wait;
	/*
	   Lock HCI before registering to prevent receiving events
	   from the same command
	*/
	if (hsk->hci->pid != current->pid)
		down_read(&hsk->hci->sema);
	hsk->lock = hci_lock(hsk->hci, hsk->opcode);
	if (!hsk->lock) {
		kfree_skb(skb);
		//if (hsk->hci->pid != current->pid)
			up_read(&hsk->hci->sema);// it will never fail on exclusive lock
		return -ENOMEM;
	}
	/* set some fields */
	if (!(flags & HCI_SKIP_STATUS))
		hsk->flags |= AFFIX_FLAGS_W4_STATUS;
	else
		hsk->flags &= ~AFFIX_FLAGS_W4_STATUS;

	if (hsk->flags & AFFIX_FLAGS_SUPER)
		atomic_inc(&hsk->hci->cmd_count);	// allow certainly to send a command

	skb_queue_purge(&sk->receive_queue);	// Purge unexpected events
	hsk->opcode = __get_u16(skb->data);	/* BT byte order */
	hsk->pkt_mask = 1 << HCI_EVENT;
	/* send a command */
	skb->pkt_type = HCI_COMMAND;
	if (!(flags & HCI_NO_UART_ENCAP))
		skb_pull(skb, 1);
	err = hci_queue_xmit(hsk->hci, NULL, skb);
	if (err) {
		kfree_skb(skb);
		goto exit;
	}
	hsk->flags |= AFFIX_FLAGS_CMD_PENDING;
wait:
	err = hpf_wait_for_completion(sk);
	if (err < 0)
		goto exit;
	// err has COMMAND_STATUS in this case
exit:
	hsk->flags &= ~AFFIX_FLAGS_CMD_PENDING;
	hpf_unlock(hsk);
	if (hsk->hci->pid != current->pid)
		up_read(&hsk->hci->sema);
	hsk->opcode = 0;	/* zero opcode */
	return err;
}

/*
    size - size of the iovec array, size of the message
*/
int hpf_sendmsg_raw(struct socket *sock, struct msghdr *msg, int size, struct scm_cookie *scm)
{
	struct sock	*sk = sock->sk;
	hci_sock_t	*hsk = hpf_get(sk);
	int		err;
	struct sk_buff	*skb = NULL;
	hci_struct	*hci = hsk->hci;

	DBFENTER;

	/* deal with restarts */
	if (hsk->flags & AFFIX_FLAGS_CMD_PENDING)
		goto wait;
	if (!hsk->hci)
		return -ENODEV;
	
	skb = sock_alloc_send_skb(sk, hsk->hci->hdrlen + size, msg->msg_flags & MSG_DONTWAIT, &err);
	if (skb == NULL)
		return err;
	skb_reserve(skb, hsk->hci->hdrlen);

	err = memcpy_fromiovec(skb_put(skb, size), msg->msg_iov, size);
	if (err)
		goto out;
		
	if (msg->msg_flags & HCI_REQUEST_MODE) {
wait:
		err = hpf_send_request(sk, skb, msg->msg_flags);
	} else {
		skb->pkt_type = *skb->data;	/*  get packet type */
		skb_pull(skb, 1);
		err = hci_queue_xmit(hci, NULL, skb);
		if (err)
			goto out;
		err = size;
	}
	DBFEXIT;
	return err;
out:
	kfree_skb(skb);
	DBPRT("err: %d\n", err);
	return err;
}

int hpf_sendmsg(struct socket *sock, struct msghdr *msg, int size, struct scm_cookie *scm)
{
	struct sock		*sk = sock->sk;
	hci_sock_t		*hsk;
	int			err = -ENOTCONN;
	struct sk_buff		*skb;

	DBFENTER;

	if (msg->msg_flags & ~MSG_DONTWAIT)
		return -EINVAL;
	if (sock->state != SS_CONNECTED) 
		return -ENOTCONN;
	if (sk->shutdown & SHUTDOWN_MASK) {
		send_sig(SIGPIPE, current, 0);
		return -EPIPE;
	}
	if (sk->state != CON_ESTABLISHED)
		return -ENOTCONN;

	hsk = hpf_get(sk);
	skb = sock_alloc_send_skb(sk, size + HCI_SKB_RESERVE, msg->msg_flags & MSG_DONTWAIT, &err);
	if (skb == NULL)
		return err;
	skb_reserve(skb, HCI_SKB_RESERVE);
	
	err = memcpy_fromiovec(skb_put(skb, size), msg->msg_iov, size);
	if (err)
		goto err;
	/* send skb */
	err = lp_send_data(hsk->con, skb);
	if (err)
		goto err;
 	DBFEXIT;
	return size;
 err:
 	kfree_skb(skb);
	return err;
}

int hci_put_cmsg(struct msghdr *msg, int level, int type, int len, void *data)
{
	struct cmsghdr	*cm = (struct cmsghdr*)msg->msg_control;
	struct cmsghdr	cmhdr;
	int 		cmlen = CMSG_SPACE(len);
	int 		err;

	if (cm == NULL || msg->msg_controllen < cmlen) {
		msg->msg_flags |= MSG_CTRUNC;
		return 0;
	}
	cmhdr.cmsg_level = level;
	cmhdr.cmsg_type = type;
	cmhdr.cmsg_len = CMSG_LEN(len);
	err = -EFAULT;
	if (copy_to_user(cm, &cmhdr, sizeof cmhdr))
		goto out; 
	if (copy_to_user(CMSG_DATA(cm), data, len))
		goto out;
	msg->msg_control += cmlen;
	msg->msg_controllen -= cmlen;
	err = 0;
out:
	return err;
}

int hpf_recvmsg_raw(struct socket *sock, struct msghdr *msg, int size, int flags, struct scm_cookie *scm)
{
	struct sock		*sk = sock->sk;
	int			err = -ENOTCONN;
	struct sk_buff		*skb;

	DBFENTER;

	skb = skb_recv_datagram(sk, (flags & ~MSG_DONTWAIT) | MSG_PEEK, flags & MSG_DONTWAIT, &err);
	if (skb == NULL)
		goto exit;
	
	DBPRT("We received %d bytes\n", skb->len);
	DBDUMP(skb->data, skb->len);
	
	if (!(flags & HCI_NO_UART_ENCAP))
		*skb_push(skb, 1) = skb->pkt_type;	// add pkt_type
	
	skb->h.raw = skb->data;
	if (skb->len <= size)
		size = skb->len;
	else if (sock->type == SOCK_SEQPACKET) {
		msg->msg_flags |= MSG_TRUNC;
		if (!(flags & MSG_PEEK))
			skb_trim(skb, size);
	}
	skb_copy_datagram_iovec(skb, 0, msg->msg_iov, size);
	if (!(flags & MSG_PEEK))
		skb_pull(skb, size);
	if (!skb->len) {
		skb_unlink(skb);
		kfree_skb(skb);
	}
	if (msg->msg_name) {
		/* put address info */
		struct sockaddr_affix	*sa = msg->msg_name;
		sa->family = AF_AFFIX;
		sa->devnum = hci_skb(skb)->devnum;
		sa->port = hci_skb(skb)->outgoing;
		msg->msg_namelen = sizeof(*sa);
	}
	kfree_skb(skb);		// remove MSG_PEEK link
	err = size;
 exit:
	DBFEXIT;
	return err;
}

int hpf_recvmsg(struct socket *sock, struct msghdr *msg, int size, int flags, struct scm_cookie *scm)
{
	struct sock		*sk = sock->sk;
	int			err = -ENOTCONN;

	DBFENTER;
	if (sock->flags & __SO_ACCEPTCON)
		return(-EINVAL);
	if (sock->state != SS_CONNECTED) 
		return -ENOTCONN;
	if (sk->state != CON_ESTABLISHED)  {
		if (sk->shutdown & SHUTDOWN_MASK) {
			//send_sig(SIGPIPE, current, 0);
			return sock_error(sk);
		}
		return -ENOTCONN;
	}
	err = hpf_recvmsg_raw(sock, msg, size, flags, scm);
	DBFEXIT;
	return err;
}

/*
    HCI Protocol Family operations
*/

struct proto_ops hpf_raw_ops = {
	family:		PF_AFFIX,

	release:	hpf_release,
	bind:		sock_no_bind,//hpf_bind,
	connect:	sock_no_connect,//hpf_connect,
	socketpair:	sock_no_socketpair,
	accept:		sock_no_accept,//hpf_accept,
	getname:	hpf_getname,
	poll:		datagram_poll,
	ioctl:		hpf_ioctl,
	listen:		sock_no_listen,//hpf_listen,
	shutdown:	sock_no_shutdown,
	setsockopt:	hpf_setsockopt,
	getsockopt:	sock_no_getsockopt,
	sendmsg:	hpf_sendmsg_raw,
	recvmsg:	hpf_recvmsg_raw,
	mmap:		sock_no_mmap,
	sendpage:	sock_no_sendpage
};

/*
 * ops for ACL/SCO sockets
 */
struct proto_ops hpf_data_ops = {
	family:		PF_AFFIX,

	release:	hpf_release,
	bind:		hpf_bind,
	connect:	hpf_connect,
	socketpair:	sock_no_socketpair,
	accept:		hpf_accept,
	getname:	hpf_getname,
	poll:		datagram_poll,
	ioctl:		hpf_ioctl,
	listen:		hpf_listen,
	shutdown:	sock_no_shutdown,
	setsockopt:	hpf_setsockopt,
	getsockopt:	sock_no_getsockopt,
	sendmsg:	hpf_sendmsg,
	recvmsg:	hpf_recvmsg,
	mmap:		sock_no_mmap,
	sendpage:	sock_no_sendpage
};


int __hpf_create(struct socket *sock, int protocol)
{
	struct sock	*sk;

	DBFENTER;
	if (sock->type == SOCK_RAW && protocol != BTPROTO_HCI)
		return -ESOCKTNOSUPPORT;
	if (sock->type != SOCK_RAW && sock->type != SOCK_SEQPACKET)
		return -ESOCKTNOSUPPORT;
	sock->state = SS_UNCONNECTED;
	sk = hpf_create(PF_AFFIX, GFP_KERNEL, 1);
	if (sk == NULL)
		return -ENOMEM;
	hpf_init_data(sock, sk);	/* generic initialization */
	sk->protocol = protocol;
	if (sock->type == SOCK_RAW)
		sock->ops = &hpf_raw_ops;
	else
		sock->ops = &hpf_data_ops;
	btl_add_tail(&hci_socks, hpf_get(sk));
	MOD_INC_USE_COUNT;
	DBFEXIT;
	return 0;
}

struct net_proto_family hpf_family_ops = {
	family:	PF_AFFIX,
	create:	__hpf_create
};

static int affix_event(struct notifier_block *nb, unsigned long event, void *arg)
{
	switch (event) {
		case HCIDEV_DOWN:
			break;
		default:
			break;
	}
	return NOTIFY_DONE;
}

static struct notifier_block affix_notifier_block = {
	notifier_call:	affix_event,
};

int hpf_proc_read(char *buf, char **start, off_t offset, int len)
{
	int count = 0;

	DBFENTER;

	count += sprintf(buf+count, "HCI sockets in use: %ld, list len: %d\n",
			sock_count, hci_socks.len);

	DBFEXIT;
	return count;
}


struct proc_dir_entry	*hpf_proc;


int __init init_hpf(void)
{
	int	err;
	
	DBFENTER;
	hpf_proc = create_proc_info_entry("hci_sock_t", 0, proc_affix, hpf_proc_read);
	if (hpf_proc == NULL) {
		BTERROR("Unable to register proc fs entry\n");
		return -EINVAL;
	}
	err = affix_sock_register(&hpf_family_ops, BTPROTO_HCI);
	if (err)
		goto err1;
	err = affix_sock_register(&hpf_family_ops, BTPROTO_HCIACL);
	if (err)
		goto err2;
	err = affix_sock_register(&hpf_family_ops, BTPROTO_HCISCO);
	if (err)
		goto err3;
	affix_register_notifier(&affix_notifier_block);
	btl_head_init(&hci_socks);
	DBFEXIT;
	return 0;
err3:
	affix_sock_unregister(BTPROTO_HCIACL);
err2:
	affix_sock_unregister(BTPROTO_HCI);
err1:
	remove_proc_entry("hci_sock_t", proc_affix);
	return err;
}

void __init exit_hpf(void)
{
	DBFENTER;

	affix_unregister_notifier(&affix_notifier_block);
	affix_sock_unregister(BTPROTO_HCISCO);
	affix_sock_unregister(BTPROTO_HCIACL);
	affix_sock_unregister(BTPROTO_HCI);
	remove_proc_entry("hci_sock_t", proc_affix);

	DBFEXIT;
}

/* Exporting stuff */
EXPORT_SYMBOL(hci_open_dev);
#if defined(CONFIG_AFFIX_PAN) || defined(CONFIG_AFFIX_PAN_MODULE)
EXPORT_SYMBOL(HCI_SwitchRole);
EXPORT_SYMBOL(HCI_WriteScanEnable);
EXPORT_SYMBOL(HCI_ReadScanEnable);
EXPORT_SYMBOL(HCI_WritePageScanActivity);
EXPORT_SYMBOL(HCI_WriteInquiryScanActivity);
EXPORT_SYMBOL(HCI_WriteClassOfDevice);
EXPORT_SYMBOL(HCI_WriteLinkSupervisionTimeout);
EXPORT_SYMBOL(HCI_ReadBDAddr);
EXPORT_SYMBOL(HCI_ReadRSSI);
EXPORT_SYMBOL(HCI_GetLinkQuality);
EXPORT_SYMBOL(HCI_ReadTransmitPowerLevel);
EXPORT_SYMBOL(HCI_Inquiry);
EXPORT_SYMBOL(HCI_ChangeConnectionPacketType);
#endif

/*
 * 1. set mode mask for connecting sockets
 * 2. disconnect based on mode mask
 */

