/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: btuart_h4.c,v 1.11 2003/02/28 10:33:39 kds Exp $

   btuart_h4 - UART protocol (H4)
   
   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
                Imre Deak <ext-imre.deak@nokia.com>
*/


/* The following prevents "kernel_version" from being set in this file. */
#define __NO_VERSION__

#include <linux/config.h>

/* Module related headers, non-module drivers should not include */
#include <linux/module.h>

/* Standard driver includes */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/skbuff.h>
#include <asm/io.h>
#include <linux/serial_reg.h>

/* Local Includes */
#define	FILEBIT	DBDRV

#include <affix/btdebug.h>
#include <affix/hci.h>
#include <affix/uart.h>

int btuart_receive_data_h4(btuart_t *btuart, struct sk_buff *skb)
{
	__u8	*hci_type = (__u8*)skb->data;
	
	DBFENTER;

	DBPRT("HCI packet, len: %d\n", skb->len);
	DBDUMP(skb->data, skb->len);

	/* remove uart header */
	skb_pull(skb, 1);
	skb->pkt_type = *hci_type;
	hcidev_rx(btuart->hci, skb);	/* send to upper protocol layer */
	
	DBFEXIT;
	return 0;
}

void btuart_recv_buf_h4(btuart_t *btuart, const unsigned char *cp, int count)
{
	const unsigned char	*ptr;
	int			rlen, lmin;

	DBFENTER;

	DBPRT("We've received %d bytes\n", count);
	DBDUMP((__u8*)cp, count);

	ptr = cp;
	rlen = count;

	while (rlen) {
		if (btuart->rx_count == 0) {	/* expect new packet */
			btuart->hdr[0] = *ptr;
			ptr += 1;
			rlen -= 1;
			
			switch (btuart->hdr[0]) {
			case UART_HCI_ACL:
				btuart->hdr_len = 1 + sizeof(HCI_ACL_Packet_Header);
				break;
			case UART_HCI_SCO:
				btuart->hdr_len = 1 + sizeof(HCI_SCO_Packet_Header);
				break;
			case UART_HCI_EVENT:
				btuart->hdr_len = 1 + sizeof(HCI_Event_Packet_Header);
				break;
			default:
				BTERROR("Invalid HCI packet type %#02x\n", btuart->hdr[0]);
				btuart->hdr_len = 0;
				break;
			}
			if (btuart->hdr_len)
				btuart->rx_count = 1;
		} else if (btuart->rx_count < btuart->hdr_len) {
			lmin = btmin(btuart->hdr_len - btuart->rx_count, rlen);
			memcpy(btuart->hdr + btuart->rx_count, ptr, lmin);
			btuart->rx_count += lmin;
			ptr += lmin;
			rlen -= lmin;

			if (btuart->rx_count >= btuart->hdr_len) {
				switch (btuart->hdr[0]) {
				case UART_HCI_ACL:
					btuart->pkt_len = btuart->hdr_len +
						__btoh16(((HCI_ACL_Packet_Header*)&btuart->hdr[1])->Length);
					break;
				case UART_HCI_SCO:
					btuart->pkt_len = btuart->hdr_len +
						((HCI_SCO_Packet_Header*)&btuart->hdr[1])->Length;
					break;
				case UART_HCI_EVENT:
					btuart->pkt_len = btuart->hdr_len +
						((HCI_Event_Packet_Header*)&btuart->hdr[1])->Length;
					break;
				}

				btuart->rx_skb = dev_alloc_skb(btuart->pkt_len);
				if (!btuart->rx_skb) {
					BTERROR("No memory for skbuff\n");
					/* No free memory */
					btuart->rx_count = 0;
					return;
				}
				/* copy tlp header */
				memcpy(skb_put(btuart->rx_skb, btuart->hdr_len), btuart->hdr, btuart->hdr_len);
			}

		} else {
			if (btuart->rx_skb->len < btuart->pkt_len) {
				lmin = btmin(btuart->pkt_len - btuart->rx_skb->len, rlen);
				memcpy(skb_put(btuart->rx_skb, lmin), ptr, lmin);
				ptr += lmin;
				rlen -= lmin;
				if (btuart->rx_skb->len >= btuart->pkt_len) {
					btuart_receive_data_h4(btuart, btuart->rx_skb);
					btuart->rx_count = 0;
					btuart->rx_skb = NULL;
				}
			}
		}
	}
	DBFEXIT;
}


int btuart_enqueue_h4(btuart_t *btuart, struct sk_buff *skb)
{
	DBFENTER;
	*skb_push(skb, 1) = skb->pkt_type;
	DBDUMP(skb->data, skb->len);
	skb_queue_tail(&btuart->tx_q, skb);
	DBFEXIT;
	return 0;
}

struct sk_buff *btuart_dequeue_h4(btuart_t *btuart)
{
	return skb_dequeue(&btuart->tx_q);
}

int btuart_init_h4(btuart_t *btuart)
{
	DBFENTER;
	DBFEXIT;
	return 0;
}

void btuart_uninit_h4(btuart_t *btuart)
{
	DBFENTER;
	DBFEXIT;
}

