/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
   
   Original Author: Imre Deak <ext-imre.deak@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: btusb.c,v 1.137 2003/03/04 13:46:52 kds Exp $

   Physical protocol layer for USB Bluetooth devices

   Fixes:
   		Dmitry Kasatkin		: Audio Support, New Request Management
*/		

#define __NO_VERSION__

#include <linux/config.h>

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/fcntl.h>

#include <linux/skbuff.h>

#include <linux/usb.h>

#define FILEBIT DBDRV		// for sake of selective debug messages

#include <affix/btdebug.h>
#include <affix/hci.h>

#ifndef USB_ZERO_PACKET
#define USB_ZERO_PACKET		0
#endif

#if 1
#undef  USB_ASYNC_UNLINK
#define USB_ASYNC_UNLINK	0
#endif

#define BTUSB_CLAIM_OPEN	1


unsigned long long	isoc_count;

/* Vendor/Product codes */
#define VENDOR_ID_DIGIANSWER			0x8fd
#define DEVICE_ID_DIGIANSWER_USB_BT_DONGLE	1
#define VENDOR_ID_SILICONWAVE			0xc10
#define DEVICE_ID_SILICONWAVE_USB_BT_MODULE	0
#define VENDOR_ID_CSR				0x0a12
#define DEVICE_ID_CSR				0x01
#define VENDOR_3COM				0x506
#define DEVICE_3COM_USB_BT_MODULE		0xa0

/* Class, SubClass, and Protocol codes that describe a Bluetooth device */
#define WIRELESS_CLASS_CODE			0xe0
#define RF_SUBCLASS_CODE			0x01
#define BLUETOOTH_PROGRAMMING_PROTOCOL_CODE	0x01

/* Sizes */
/* ACL */
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 4, 8)
#define BTUSB_ACL_RX_URB		1
#else
//#define BTUSB_ACL_RX_URB		8
#define BTUSB_ACL_RX_URB		1
#endif
#define BTUSB_ACL_RX_SIZE		2048
#define BTUSB_ACL_TX_URB		4
/* SCO RX*/
#define BTUSB_SCO_RX_URB		2
#define BTUSB_SCO_RX_SIZE		32
#define BTUSB_SCO_RX_PACKETS		10
/* SCO TX */
#define BTUSB_SCO_TX_URB		2
#define BTUSB_SCO_TX_SIZE		8
#define BTUSB_SCO_TX_PACKETS		10

/*  Maximum number of consecutive failed USB reads before giving up. */
#define MAX_READ_FAIL_COUNT		100

typedef struct _usb_bt_params
{
	int	uart_leading_byte;	// if there is a uart-style leading byte for packets
	__u8	control_req_type;
	int	vendor_id;
	int	device_id;
} usb_bt_params_t;

usb_bt_params_t usb_bt_list[] =
{
	{
		uart_leading_byte : 1,
		control_req_type : 0x40,
		vendor_id : VENDOR_ID_DIGIANSWER,
		device_id : DEVICE_ID_DIGIANSWER_USB_BT_DONGLE
	},
        {       uart_leading_byte : 0,		// default last descriptor
                control_req_type : 0x20,
                vendor_id : 0,
                device_id : 0 
        }
};

#define NUM_OF_USB_BT_DEV (sizeof(usb_bt_list) / sizeof(usb_bt_list[0]))	

static struct usb_device_id btusb_ids [] = {
	/* Generic Bluetooth USB device */
	{ USB_DEVICE_INFO(WIRELESS_CLASS_CODE,
			RF_SUBCLASS_CODE, BLUETOOTH_PROGRAMMING_PROTOCOL_CODE)},

	/* Ericsson with non-standard id */
	{ USB_DEVICE(0x0bdb, 0x1002) },
	
	{ }
};

MODULE_DEVICE_TABLE (usb, btusb_ids);

static void *btusb_probe(struct usb_device *dev, unsigned int ifnum,
			 	 const struct usb_device_id *id);
static void btusb_disconnect(struct usb_device *dev, void *ptr);

static struct usb_driver btusb_driver =
{
	name:		"affix_usb",
	probe:		btusb_probe,
	disconnect:	btusb_disconnect,
	id_table:	btusb_ids
};


/* request data structure */
typedef struct btusb_req {
	struct btusb		*btusb;
#define BTUSB_REQ_WAKEUP	0
	int			flags;
	rwlock_t		callback_lock;
	spinlock_t		xmit_lock;
	unsigned int		pipe;
	__u8			interval;

	/* rx state */
	struct sk_buff		*rx_skb;
	/* tx state */
	struct sk_buff_head	queue;
	struct sk_buff_head	pending;
	
} btusb_req_t;

typedef struct {
	struct urb	*urb;	
	btusb_req_t	*req;
	int		pkt_len;
} btusb_skb_cb;

#define btusb_cb(skb)	((btusb_skb_cb*)skb->cb)

/* driver data structure */
typedef struct btusb {
	struct list_head	q;	/* for queueing */
	
	int			signature;
	struct usb_device	*usb_dev;
	hci_struct		*hci;
	usb_bt_params_t		*params;
#define BTUSB_RUNNING		0x00
#define BTUSB_AUDIO_IN		0x01
#define BTUSB_AUDIO_OUT		0x02
	int			flags;
	int			isoc;
	
	btusb_req_t		ctrl_out_req;
	btusb_req_t		int_in_req;
	btusb_req_t		bulk_in_req;
	btusb_req_t		bulk_out_req;
	btusb_req_t		isoc_in_req;
	btusb_req_t		isoc_out_req;

	int			read_fail_cnt;

} btusb_t;


static void btusb_out_complete(struct urb *urb);
static void btusb_int_in_complete(struct urb *urb);
static void btusb_bulk_in_complete(struct urb *urb);
static void btusb_isoc_in_complete(struct urb *urb);
static void btusb_isoc_out_complete(struct urb *urb);

static void btusb_init_req(btusb_req_t *req, btusb_t *btusb)
{
	req->btusb = btusb;
	rwlock_init(&req->callback_lock);
	spin_lock_init(&req->xmit_lock);
	skb_queue_head_init(&req->queue);
	skb_queue_head_init(&req->pending);
	req->rx_skb = NULL;
}

static void btusb_release_req(btusb_req_t *req)
{
	struct sk_buff	*skb;
	struct urb	*urb;

	DBFENTER;
	write_lock(&req->callback_lock);
	/* purge data queue */
	while ((skb = skb_dequeue(&req->queue)))
		dev_kfree_skb_any(skb);
	/* purge urb queue */
	while ((skb = skb_dequeue(&req->pending))) {
		urb = btusb_cb(skb)->urb;
		usb_unlink_urb(urb);
		if (usb_pipecontrol(urb->pipe))
			kfree(urb->setup_packet);
		if (!USB_ASYNC_UNLINK)
			usb_free_urb(urb);	//FIXME: necessary???
		dev_kfree_skb_any(skb);		
	}
	if (req->rx_skb) {
		dev_kfree_skb_any(req->rx_skb);
		req->rx_skb = NULL;
	}
	write_unlock(&req->callback_lock);
	DBFEXIT;
}


static inline int btusb_submit_urb(struct urb *urb)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	return usb_submit_urb(urb);
#else
	return usb_submit_urb(urb, GFP_ATOMIC);
#endif
}


static inline int __usb_fill_isoc_urb(struct urb *urb, struct usb_device *dev,
		unsigned int pipe, struct sk_buff *skb, int num_packets, int mtu,
		usb_complete_t complete, void *context) 
{
	int 	k, len = skb->len, offset = 0;

	spin_lock_init(&urb->lock);
	urb->dev = dev;
	urb->pipe = pipe;
	urb->transfer_buffer = skb->data;
	urb->number_of_packets = num_packets;
	urb->actual_length = 0;
	urb->complete = complete;
	urb->context = context;
	urb->transfer_flags = USB_ISO_ASAP;
	for (k = 0; k < num_packets && len; k++) {
		urb->iso_frame_desc[k].offset = offset;
		urb->iso_frame_desc[k].actual_length = 0;
		urb->iso_frame_desc[k].length = (len>mtu)?mtu:len;
		len -= urb->iso_frame_desc[k].length;
		offset += urb->iso_frame_desc[k].length;
	}
	urb->number_of_packets = k;
	urb->transfer_buffer_length = skb->len - len;
	return urb->transfer_buffer_length;
}

static inline void btusb_fill_isoc_urb(struct urb *urb, struct usb_device *dev,
		unsigned int pipe, struct sk_buff *skb, int num_packets, int mtu,
		usb_complete_t complete, void *context) 
{
	int 	len;

	len = __usb_fill_isoc_urb(urb, dev, pipe, skb, num_packets, mtu, complete, context);
	skb_pull(skb, len);
}

int btusb_setup_isoc_in(btusb_req_t *req)
{
	int		err;
	struct urb	*urb;
	struct sk_buff	*skb;
	int		mtu, num_packets;

#ifdef BTUSB_SCO_RX_SIZE
	mtu = BTUSB_SCO_RX_SIZE;
#else
	mtu = usb_maxpacket(req->btusb->usb_dev, req->pipe, 0);
#endif
	if (mtu <= 0)
		return -EINVAL;
	num_packets = BTUSB_SCO_RX_PACKETS;
	
	DBFENTER;

	urb = usb_alloc_urb(num_packets);
	if (!urb)
		return -ENOMEM;

	skb = dev_alloc_skb(num_packets * mtu);
	if (!skb) {
		usb_free_urb(urb);
		return -ENOMEM;
	}
	skb_put(skb, num_packets * mtu);

	__usb_fill_isoc_urb(urb, req->btusb->usb_dev, req->pipe, 
			skb, num_packets, mtu,
			btusb_isoc_in_complete, skb);
	btusb_cb(skb)->urb = urb;
	btusb_cb(skb)->req = req;

	skb_queue_tail(&req->pending, skb);
	err = btusb_submit_urb(urb);
	if (err) {
		skb_unlink(skb);
		dev_kfree_skb_any(skb);
		usb_free_urb(urb);
	}
	DBFEXIT;
	return err;
}

int btusb_audio_mode(btusb_t *btusb, int audio_mode)
{
	int	err, i;

	DBFENTER;

	if (!btusb->isoc)
		return -ENODEV;
	if (test_bit(BTUSB_RUNNING, &btusb->flags)) {
		DBPRT("usb_set_interface(): %d\n", AFFIX_AUDIO_GETALT(audio_mode));
		err = usb_set_interface(btusb->usb_dev, 1, AFFIX_AUDIO_GETALT(audio_mode));
		if (err)
			return err;
	}
	if (audio_mode & AFFIX_AUDIO_IN) { 
		if (!test_and_set_bit(BTUSB_AUDIO_IN, &btusb->flags)) {
			isoc_count = 0;
			for (i = 0; i < BTUSB_SCO_RX_URB; i++) {
				err = btusb_setup_isoc_in(&btusb->isoc_in_req);
				if (err) {
					BTERROR("setup isoc pipes failed with status %d\n", err);
					clear_bit(BTUSB_AUDIO_IN, &btusb->flags);
					btusb_release_req(&btusb->isoc_in_req);
					return err;
				}
			}
		}
	} else {
		if (test_and_clear_bit(BTUSB_AUDIO_IN, &btusb->flags)) {
			btusb_release_req(&btusb->isoc_in_req);
			BTINFO("Received isoc: %llu\n", isoc_count);
			isoc_count = 0;
		}
	}
	if (audio_mode & AFFIX_AUDIO_OUT) {
		if (!test_and_set_bit(BTUSB_AUDIO_OUT, &btusb->flags)) {
			/* nothing */;
		}
	} else {
		if (test_and_clear_bit(BTUSB_AUDIO_OUT, &btusb->flags)) {
			btusb_release_req(&btusb->isoc_out_req);
		}
	}
	DBFEXIT;
	return 0;
}


int btusb_setup_int_in(btusb_req_t *req)
{
	int		err;
	struct urb	*urb;
	struct sk_buff	*skb;
	int		mtu;

	DBFENTER;

	mtu = usb_maxpacket(req->btusb->usb_dev, req->pipe, 0);
	if (mtu <= 0)
		return -EINVAL;
	
	urb = usb_alloc_urb(0);
	if (!urb)
		return -ENOMEM;

	skb = dev_alloc_skb(mtu);
	if (!skb) {
		usb_free_urb(urb);
		return -ENOMEM;
	}
	skb_put(skb, mtu);

	FILL_INT_URB(urb, req->btusb->usb_dev, req->pipe,
			skb->data, skb->len,
			btusb_int_in_complete, skb, req->interval);
	btusb_cb(skb)->urb = urb;
	btusb_cb(skb)->req = req;

	skb_queue_tail(&req->pending, skb);
	err = btusb_submit_urb(urb);
	if (err) {
		skb_unlink(skb);
		dev_kfree_skb_any(skb);
		usb_free_urb(urb);
	}
	
	DBFEXIT;
	return err;
}

int btusb_setup_bulk_in(btusb_req_t *req)
{
	int		err;
	struct urb	*urb;
	struct sk_buff	*skb;
	int		mtu;

	DBFENTER;
	
#if BTUSB_ACL_RX_URB > 2
	mtu = usb_maxpacket(req->btusb->usb_dev, req->pipe, 0);
	if (mtu <= 0)
		return -EINVAL;
#else
	mtu = BTUSB_ACL_RX_SIZE;
#endif

	urb = usb_alloc_urb(0);
	if (!urb)
		return -ENOMEM;

	skb = dev_alloc_skb(mtu);
	if (!skb) {
		usb_free_urb(urb);
		return -ENOMEM;
	}
	skb_put(skb, mtu);

	FILL_BULK_URB(urb, req->btusb->usb_dev, req->pipe,
			skb->data, skb->len,
			btusb_bulk_in_complete, skb);
	urb->transfer_flags = USB_QUEUE_BULK | USB_ASYNC_UNLINK;
	btusb_cb(skb)->urb = urb;
	btusb_cb(skb)->req = req;

	skb_queue_tail(&req->pending, skb);
	err = btusb_submit_urb(urb);
	if (err) {
		skb_unlink(skb);
		dev_kfree_skb_any(skb);
		usb_free_urb(urb);
	}
	DBFEXIT;
	return err;
}

int btusb_xmit_bulk(btusb_req_t *req, struct sk_buff *skb)
{
	int		err;
	struct urb	*urb;

	DBFENTER;

	if (!test_bit(BTUSB_RUNNING, &req->btusb->flags))
		return -1;

	urb = usb_alloc_urb(0);
	if (!urb)
		return -ENOMEM;

	FILL_BULK_URB(urb, req->btusb->usb_dev, req->pipe,
			skb->data, skb->len,
			btusb_out_complete, skb);
	urb->transfer_flags = USB_QUEUE_BULK | USB_ZERO_PACKET | USB_ASYNC_UNLINK;
	btusb_cb(skb)->urb = urb;
	btusb_cb(skb)->req = req;

	skb_queue_tail(&req->pending, skb);
	err = btusb_submit_urb(urb);
	if (err) {
		skb_unlink(skb);
		usb_free_urb(urb);
	}
	
	DBFEXIT;
	return err;
}

int btusb_xmit_isoc(btusb_req_t *req, struct sk_buff *skb)
{
	int		err;
	struct urb	*urb;
	int		mtu, num_packets;

	DBFENTER;
	
	if (!test_bit(BTUSB_RUNNING, &req->btusb->flags))
		return -1;

#ifdef BTUSB_SCO_TX_SIZE
	mtu = BTUSB_SCO_TX_SIZE;
#else
	mtu = usb_maxpacket(req->btusb->usb_dev, req->pipe, 1);
#endif
	if (mtu <= 0)
		return -EINVAL;
	num_packets = BTUSB_SCO_TX_PACKETS;

	urb = usb_alloc_urb(num_packets);
	if (!urb)
		return -ENOMEM;
	
	btusb_fill_isoc_urb(urb, req->btusb->usb_dev, req->pipe, 
			skb, num_packets, mtu,
			btusb_isoc_out_complete, skb);
	btusb_cb(skb)->urb = urb;
	btusb_cb(skb)->req = req;

	skb_queue_tail(&req->pending, skb);
	err = btusb_submit_urb(urb);
	if (err) {
		skb_unlink(skb);
		usb_free_urb(urb);
	}
	DBFEXIT;
	return err;
}

int __btusb_xmit_isoc(btusb_req_t *req, struct sk_buff *src_skb)
{
	int		err;
	struct urb	*urb;
	struct sk_buff	*skb;
	int		num_packets, num, len, mtu;

	DBFENTER;

#ifdef BTUSB_SCO_TX_SIZE
	mtu = BTUSB_SCO_TX_SIZE;
#else
	mtu = usb_maxpacket(req->btusb->usb_dev, req->pipe, 1);
#endif
	if (mtu <= 0)
		return -EINVAL;
	num_packets = BTUSB_SCO_TX_PACKETS;
	
	num = src_skb->len / mtu;
	if (src_skb->len % mtu)
		num++;
	if (num < num_packets)
		num_packets = num;

	urb = usb_alloc_urb(num_packets);
	if (!urb)
		return -ENOMEM;

	skb = dev_alloc_skb(num_packets * mtu);
	if (!skb) {
		usb_free_urb(urb);
		return -ENOMEM;
	}
	skb_put(skb, btmin(num_packets * mtu, src_skb->len));

	len = __usb_fill_isoc_urb(urb, req->btusb->usb_dev, req->pipe, 
			skb, num_packets, mtu,
			btusb_isoc_out_complete, skb);
	btusb_cb(skb)->urb = urb;
	btusb_cb(skb)->req = req;

	memcpy(skb->data, src_skb->data, len);
	skb_pull(src_skb, len);

	DBPRT("btusb_submit_urb(), len: %d, left: %d\n", len, src_skb->len);
	skb_queue_tail(&req->pending, skb);
	err = btusb_submit_urb(urb);
	if (err) {
		skb_unlink(skb);
		dev_kfree_skb_any(skb);
		usb_free_urb(urb);
	}
	DBFEXIT;
	return err;
}


int btusb_xmit_ctrl(btusb_req_t *req, struct sk_buff *skb)
{
	int		err;
	struct urb	*urb;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 20)
	devrequest	*dr;		/* for setup packets */
#else
	struct usb_ctrlrequest	*dr;		/* for setup packets */
#endif

	DBFENTER;

	if (!test_bit(BTUSB_RUNNING, &req->btusb->flags))
		return -1;

	urb = usb_alloc_urb(0);
	if (!urb)
		return -ENOMEM;

	dr = kmalloc(sizeof(*dr), GFP_ATOMIC);
	if (!dr) {
		usb_free_urb(urb);
		return -ENOMEM;
	}
	memset(dr, 0, sizeof(*dr));

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 20)
	dr->requesttype = req->btusb->params->control_req_type;
	dr->length = cpu_to_le16(skb->len);
#else
	dr->bRequestType = req->btusb->params->control_req_type;
	dr->wLength = cpu_to_le16(skb->len);
#endif

	FILL_CONTROL_URB(urb, req->btusb->usb_dev, req->pipe,
			(unsigned char*)dr, skb->data, skb->len, 
			btusb_out_complete, skb);
	btusb_cb(skb)->urb = urb;
	btusb_cb(skb)->req = req;

	skb_queue_tail(&req->pending, skb);
	err = btusb_submit_urb(urb);
	if (err) {
		skb_unlink(skb);
		kfree(dr);
		usb_free_urb(urb);
	}
	
	DBFEXIT;
	return err;
}

/* it's called under spinlock */
int btusb_xmit_start(btusb_req_t *req)
{
	int		err = 0;
	struct sk_buff	*skb = NULL;

	DBFENTER;

	if (usb_pipebulk(req->pipe)) {
		while (!err && skb_queue_len(&req->pending) <= BTUSB_ACL_TX_URB && 
				(skb = skb_dequeue(&req->queue))) {
			err = btusb_xmit_bulk(req, skb);
		}
	} else if (usb_pipecontrol(req->pipe)) {
		if (skb_queue_empty(&req->pending) &&
				(skb = skb_dequeue(&req->queue)))
			err = btusb_xmit_ctrl(req, skb);
	} else {
		DBPRT("isoc_out_req, queue len: %d\n", skb_queue_len(&req->pending));
		while (!err && skb_queue_len(&req->pending) < BTUSB_SCO_TX_URB && 
				(skb = skb_dequeue(&req->queue))) {
			err = __btusb_xmit_isoc(req, skb);
#if 1			
			if (err)
				break;
			if (skb->len)
				skb_queue_head(&req->queue, skb);
			else
				dev_kfree_skb_any(skb);
#endif
		}
	}
	if (err) {
		BTERROR("usb_submit_urb failed (%d)\n", err);
		skb_queue_head(&req->queue, skb);
	} else if (skb_queue_empty(&req->queue))
		hcidev_wake_queue(req->btusb->hci);

	DBFEXIT;
	return err;
}

int btusb_xmit_wakeup(btusb_req_t *req)
{
	int	err = 0;

	DBFENTER;
	set_bit(BTUSB_REQ_WAKEUP, &req->flags);
restart:
	if (spin_trylock_bh(&req->xmit_lock)) {
		clear_bit(BTUSB_REQ_WAKEUP, &req->flags);
		err = btusb_xmit_start(req);
		spin_unlock_bh(&req->xmit_lock);
		if (!err && test_bit(BTUSB_REQ_WAKEUP, &req->flags))
			goto restart;
	}
	DBFEXIT;
	return err;
}


/*----------------------------------------------------------------------------------------------
				USB callback functions
  ----------------------------------------------------------------------------------------------
*/
static inline void btusb_receive_skb(btusb_t *btusb, btusb_req_t *req, unsigned int pkt_type)
{
	struct sk_buff *skb = req->rx_skb;

	DBPRT("new HCI pkt received. type=%#.2x  len=%d\n", pkt_type, skb->len);
	skb->pkt_type = pkt_type;
	hcidev_rx(btusb->hci, skb);
	req->rx_skb = NULL;
}

static inline btusb_t *get_btusb(btusb_req_t *req, const char *function)
{
	if (req == NULL || req->btusb == NULL || req->btusb->signature != 0xDEADBEAF) {
		BTERROR("%s - invalid btusb\n", function);
		return NULL;
	}
	return req->btusb;
}


static void btusb_out_complete(struct urb *urb)
{
	struct sk_buff	*skb = urb->context;
	btusb_req_t	*req = btusb_cb(skb)->req;
	btusb_t		*btusb = get_btusb(req, __FUNCTION__);

	DBFENTER;

	if (!test_bit(BTUSB_RUNNING, &btusb->flags))
		return;

	read_lock(&req->callback_lock);

	if (urb->status) {
		BTDEBUG("nonzero write status: %d\n", urb->status);
	}
	/* clear state */
	if (usb_pipecontrol(urb->pipe)) {
		kfree(urb->setup_packet);
	}
	skb_unlink(skb);
	dev_kfree_skb_any(skb);
	usb_free_urb(urb);
	btusb_xmit_wakeup(req);

	read_unlock(&req->callback_lock);
	DBFEXIT;
}


static void btusb_int_in_complete(struct urb *urb)
{
	struct sk_buff	*skb = urb->context;
	btusb_req_t	*req = btusb_cb(skb)->req;
	btusb_t		*btusb = get_btusb(req, __FUNCTION__);
	int 		len = urb->actual_length;
	void		*dst;
	void		*src;

	DBFENTER;
	
	if (!test_bit(BTUSB_RUNNING, &btusb->flags))
		return;

	read_lock(&req->callback_lock);
	
	if (urb->status) {
		BTDEBUG("nonzero int status received: %d\n", urb->status);
		if (++btusb->read_fail_cnt > MAX_READ_FAIL_COUNT) {
			BTDEBUG("Maximum read failure count reached. Give up.\n");
		}
		goto exit;
	}

	btusb->read_fail_cnt = 0;

	if (!len) {
		DBPRT("zero length interrupt pkt\n");
		goto exit;
	}

	DBPRT("interrupt pkt. len=%d\n", len);
	DBDUMP(urb->transfer_buffer, len);
	
	if (req->rx_skb == NULL) {	/* new packet */
		int	pkt_len;
		len -= btusb->params->uart_leading_byte;
		if (len < HCI_EVENT_HDR_LEN) {
			BTDEBUG("too small int packet\n");
			goto exit;
		}
		pkt_len = ((unsigned char *)urb->transfer_buffer)
					[1 + btusb->params->uart_leading_byte]
				 + HCI_EVENT_HDR_LEN;
		req->rx_skb = dev_alloc_skb(pkt_len);
		if (req->rx_skb == NULL) {
			BTERROR("dev_alloc_skb failed\n");
			goto exit;
		}
		btusb_cb(req->rx_skb)->pkt_len = pkt_len;
		src = (char *)urb->transfer_buffer + btusb->params->uart_leading_byte;
	} else
		src = urb->transfer_buffer;
	
	if (req->rx_skb->len + len > btusb_cb(req->rx_skb)->pkt_len) {
		BTDEBUG("USB pkt too big\n");
		dev_kfree_skb_any(req->rx_skb);
		req->rx_skb = NULL;
		goto exit;
	}
	dst = skb_put(req->rx_skb, len);
	memcpy(dst, src, len);
	if (req->rx_skb->len == btusb_cb(req->rx_skb)->pkt_len)
		btusb_receive_skb(btusb, req, HCI_EVENT);	/*  pass skb to network layer above us */
exit:
	read_unlock(&req->callback_lock);
	DBFEXIT;
}

static void btusb_bulk_in_complete(struct urb *urb)
{
	struct sk_buff	*skb = urb->context;
	btusb_req_t	*req = btusb_cb(skb)->req;
	btusb_t		*btusb = get_btusb(req, __FUNCTION__);
	unsigned char	*dst, *src;
	int	 	len = urb->actual_length;

	DBFENTER;

	if (!test_bit(BTUSB_RUNNING, &btusb->flags))
		return;

	read_lock(&req->callback_lock);

	if (urb->status) {
		BTDEBUG("nonzero read status received: %d\n", urb->status);
		if (++btusb->read_fail_cnt > MAX_READ_FAIL_COUNT) {
			BTDEBUG("Maximum read failure count reached. Give up.\n");
			goto exit;
		}
		goto resubmit;
	}
	btusb->read_fail_cnt = 0;

	if (!len) {
		//DBPRT("zero length read\n");
		goto resubmit;
	}

	DBPRT("bulk in pkt. len=%d\n", len);
	DBDUMP(urb->transfer_buffer, len);
	src = (unsigned char *)urb->transfer_buffer;
	while (len) {	
		int	copy_bytes;
		
		if (req->rx_skb == NULL) {	/* new packet */
			int	pkt_len;
			if (btusb->params->uart_leading_byte) {
				len--;
				src++;
			}
			if (len < HCI_ACL_HDR_LEN) {
				BTDEBUG("too small read packet\n");
				goto resubmit;
			}
			pkt_len = __le16_to_cpu(*(unsigned short *)(&(src[2]))) + HCI_ACL_HDR_LEN;
			req->rx_skb = dev_alloc_skb(pkt_len);
			if (req->rx_skb == NULL) {
				BTERROR("dev_alloc_skb() failed\n");
				goto resubmit;
			}
			btusb_cb(req->rx_skb)->pkt_len = pkt_len;
		}
		copy_bytes = btmin(btusb_cb(req->rx_skb)->pkt_len - req->rx_skb->len, len);
		dst = skb_put(req->rx_skb, copy_bytes);
		memcpy(dst, src, copy_bytes);
		if (req->rx_skb->len == btusb_cb(req->rx_skb)->pkt_len)
			btusb_receive_skb(btusb, req, HCI_ACL);	/*  pass skb to network layer above us */
		src += copy_bytes;
		len -= copy_bytes;
	}			
resubmit:
	urb->dev = btusb->usb_dev;
	if (btusb_submit_urb(urb)) {
		BTERROR("bulk_in resubmit failed\n");
		skb_unlink(skb);
		dev_kfree_skb_any(skb);
		usb_free_urb(urb);
	}
exit:
	read_unlock(&req->callback_lock);
	DBFEXIT;
}

/*
 * AUDIO
 */
static void btusb_isoc_out_complete(struct urb *urb)
{
	struct sk_buff	*skb = urb->context;
	btusb_req_t	*req = btusb_cb(skb)->req;
	btusb_t		*btusb = get_btusb(req, __FUNCTION__);
	int	 	len = urb->actual_length;

	if (len == 0) {
		//DBPRT("zero length written\n");
		return;
	}
	
	DBFENTER;

	if (!test_bit(BTUSB_RUNNING, &btusb->flags) || 
			!test_bit(BTUSB_AUDIO_OUT, &btusb->flags))
		return;

	read_lock(&req->callback_lock);
	
	DBPRT("wrote: %d, status: %d\n", len, urb->status);
	if (urb->status) {
		BTDEBUG("nonzero write status: %d\n", urb->status);
	}

	skb_unlink(skb);
	dev_kfree_skb_any(skb);
	usb_free_urb(urb);
	btusb_xmit_wakeup(req);

	read_unlock(&req->callback_lock);
	DBFEXIT;
}

static void btusb_isoc_in_complete(struct urb *urb)
{
	struct sk_buff	*skb = urb->context;
	btusb_req_t	*req = btusb_cb(skb)->req;
	btusb_t		*btusb = get_btusb(req, __FUNCTION__);
	unsigned char	*dst, *src;
	int	 	len = urb->actual_length, i;

	if (len) DBFENTER;

	if (!test_bit(BTUSB_RUNNING, &btusb->flags) || 
			!test_bit(BTUSB_AUDIO_IN, &btusb->flags))
		return;

	read_lock(&req->callback_lock);

	if (urb->status) {
		DBPRT("nonzero read status received: %d\n", urb->status);
		if (++btusb->read_fail_cnt > MAX_READ_FAIL_COUNT) {
			BTERROR("Maximum read failure count reached. Give up.\n");
			goto exit;
		}
		goto resubmit;
	}
	btusb->read_fail_cnt = 0;

	if (!len) {
		//DBPRT("zero length read\n");
		goto resubmit;
	}
	
	DBPRT("isoc in pkt. total len=%d\n", len);
	for (i = 0; i < urb->number_of_packets; i++) {
		if (urb->iso_frame_desc[i].status) {
			BTDEBUG("frame %u status %d\n", i, urb->iso_frame_desc[i].status);
			continue;
		}
		src = (unsigned char *)urb->transfer_buffer + urb->iso_frame_desc[i].offset;
		len = urb->iso_frame_desc[i].actual_length;
		
		urb->iso_frame_desc[i].actual_length = 0;
		urb->iso_frame_desc[i].status = 0;

		isoc_count += len;

		DBPRT("isoc in pkt. len=%d\n", len);
		DBDUMP(urb->transfer_buffer, len);

		while (len) {	
			int	copy_bytes;

			if (req->rx_skb == NULL) {	/* new packet */
				int	pkt_len;
				if (btusb->params->uart_leading_byte) {
					len--;
					src++;
				}
				if (len < HCI_SCO_HDR_LEN) {
					BTDEBUG("too small read packet\n");
					goto resubmit;
				}
				pkt_len = *(unsigned char *)(&(src[2])) + HCI_SCO_HDR_LEN;
				req->rx_skb = dev_alloc_skb(pkt_len);
				if (req->rx_skb == NULL) {
					BTERROR("dev_alloc_skb() failed\n");
					goto resubmit;
				}
				btusb_cb(req->rx_skb)->pkt_len = pkt_len;
			}
			copy_bytes = btmin(btusb_cb(req->rx_skb)->pkt_len - req->rx_skb->len, len);
			dst = skb_put(req->rx_skb, copy_bytes);
			memcpy(dst, src, copy_bytes);
			if (req->rx_skb->len == btusb_cb(req->rx_skb)->pkt_len)
				btusb_receive_skb(btusb, req, HCI_SCO);	/*  pass skb to network layer above us */
			src += copy_bytes;
			len -= copy_bytes;
		}
	}
resubmit:
	urb->dev = btusb->usb_dev;
	urb->actual_length = 0;
	if (btusb_submit_urb(urb)) {
		BTERROR("isoc_in resubmit failed\n");
		skb_unlink(skb);
		dev_kfree_skb_any(skb);
		usb_free_urb(urb);
	}
exit:
	read_unlock(&req->callback_lock);
	if (len) DBFEXIT;
}

/*----------------------------------------------------------------------------------------------
				Network interface functions.
  ----------------------------------------------------------------------------------------------
*/

#ifdef BTUSB_CLAIM_OPEN

int btusb_claim_interfaces(btusb_t *btusb)
{
	struct usb_interface	*iface = &btusb->usb_dev->actconfig->interface[0];
	int			err = 0;

	DBFENTER;

	if (usb_interface_claimed(iface))
		return -EBUSY;

	usb_driver_claim_interface(&btusb_driver, iface, btusb);

	if (!btusb->isoc)
		return 0;

	iface++;	/* [1] */

	if (usb_interface_claimed(iface)) {
		usb_driver_release_interface(&btusb_driver, iface-1);
		return -EBUSY;
	}
	usb_driver_claim_interface(&btusb_driver, iface, btusb);
#if 0
	err = usb_set_interface(btusb->usb_dev, 1, 2);
	if (err) {
		BTNOTICE("usb_set_interface(dev, 1, 2) failed");
		err = 0;
		//usb_driver_release_interface(&btusb_driver, iface);
		//usb_driver_release_interface(&btusb_driver, iface-1);
	}
#endif
	DBFEXIT;
	return err;
}

void btusb_release_interfaces(btusb_t *btusb)
{
	struct usb_interface	*iface = &btusb->usb_dev->actconfig->interface[0];

	DBFENTER;
	usb_driver_release_interface(&btusb_driver, iface++);
	if (btusb->isoc)
		usb_driver_release_interface(&btusb_driver, iface);
	DBFEXIT;
}

#endif	/* BTUSB_CLAIM_OPEN */

static int btusb_net_open(hci_struct *hci)
{
	btusb_t		*btusb = hci->priv;
	int		err, i;
	//unsigned long	flags;

	DBFENTER;

	if (test_and_set_bit(BTUSB_RUNNING, &btusb->flags)) {
		BTERROR("device already open\n");
		return -EBUSY;
	}

#ifdef BTUSB_CLAIM_OPEN
	err = btusb_claim_interfaces(btusb);
	if (err) {
		BTERROR("unable to claim interfaces: %d\n", err);
		clear_bit(BTUSB_RUNNING, &btusb->flags);
		return err;
	}
#endif
	//local_irq_save(flags);
	err = btusb_setup_int_in(&btusb->int_in_req);
	if (err) {
		BTERROR("int setup failed: %d\n", err);
		goto err1;
	}
	for (i = 0; i < BTUSB_ACL_RX_URB; i++) {
		err = btusb_setup_bulk_in(&btusb->bulk_in_req);
		if (err) {
			BTERROR("bulk setup failed: %d\n", err);
			goto err1;
		}
	}
	//local_irq_restore(flags);
	hcidev_start_queue(hci);
	DBFEXIT;
	return 0;
err1:
	clear_bit(BTUSB_RUNNING, &btusb->flags);
	btusb_release_req(&btusb->bulk_in_req);
	btusb_release_req(&btusb->int_in_req);
	//local_irq_restore(flags);
#ifdef BTUSB_CLAIM_OPEN
	btusb_release_interfaces(btusb);
#endif
	return err;
}

static int btusb_net_stop(hci_struct *hci)
{
	btusb_t 	*btusb = hci->priv;
	//unsigned long	flags;

	DBFENTER;

	if (!test_and_clear_bit(BTUSB_RUNNING, &btusb->flags))
		return 0;

	hcidev_stop_queue(hci);

	/* switch off audio */
	btusb_audio_mode(btusb, 0);

	/* unlinking urb means calling the completion routine of any pending transaction
	   it's not a problem since there we will check btusb->active.
	 */
	//local_irq_save(flags);
	btusb_release_req(&btusb->int_in_req);
	btusb_release_req(&btusb->ctrl_out_req);
	btusb_release_req(&btusb->bulk_in_req);
	btusb_release_req(&btusb->bulk_out_req);
	//local_irq_restore(flags);

#ifdef BTUSB_CLAIM_OPEN
	btusb_release_interfaces(btusb);
#endif
	DBFEXIT;
	return 0;
}

static int btusb_net_xmit_start(hci_struct *hci, struct sk_buff *skb)
{
	btusb_t		*btusb = hci->priv;
	btusb_req_t	*req = NULL;
	int		res = 0;

	DBFENTER;

	if (!hcidev_running(hci)) {
		BTERROR("%s: xmit called when interface is down\n", hci->name);
		dev_kfree_skb(skb);
		goto exit;
	}

	if (!test_bit(BTUSB_RUNNING, &btusb->flags)) {
		BTERROR("device not opened\n");
		res = -EINVAL;
		goto exit;
	}
	
	DBDUMP(skb->data, skb->len);

	if (btusb->params->uart_leading_byte)
		*skb_push(skb, 1) = skb->pkt_type;

	switch (skb->pkt_type) {
		case HCI_COMMAND:
			DBPRT("send HCI CMD pkt. len=%d\n", skb->len);
			req = &btusb->ctrl_out_req;
			skb_queue_tail(&req->queue, skb);
			break;
		case HCI_ACL:
			DBPRT("send HCI ACL pkt. len=%d\n", skb->len);
			req = &btusb->bulk_out_req;
			skb_queue_tail(&req->queue, skb);
			break;
		case HCI_SCO:
			if (!test_bit(BTUSB_AUDIO_OUT, &btusb->flags)) {
				BTERROR("Audio OUT disabled\n");
				res = -EINVAL;
				goto exit;
			}
			DBPRT("send HCI SCO pkt. len=%d\n", skb->len);
			req = &btusb->isoc_out_req;
			skb_queue_tail(&req->queue, skb);
			break;
		default:
			BTERROR("invalid HCI packet type: %d\n", skb->pkt_type);
			dev_kfree_skb_any(skb);
			goto exit;
	}
	hci->trans_start = jiffies;
	btusb_xmit_wakeup(req);
exit:
	DBFEXIT;
	return res;
} 

static int btusb_net_ioctl(hci_struct *hci, int cmd, void *arg)
{
        btusb_t		*btusb = hci->priv;
        int		res = 0;

	DBFENTER;

	if (btusb == NULL || btusb->signature != 0xDEADBEAF) {
		BTERROR("invalid btusb\n");
		res = -ENODEV;
		goto exit;
	} 
	switch (cmd) {
		case SIOCHCI_SET_AUDIO:
			res = btusb_audio_mode(btusb, hci->audio_mode);
			break;
		default:
			res = -ENOIOCTLCMD;
			goto exit;
	}
exit: 
 	DBFEXIT;
        return res;
}

/*----------------------------------------------------------------------------------------------
				USB init and shutdown functions
  ----------------------------------------------------------------------------------------------
*/

#ifdef	BTUSB_CLAIM_OPEN

int btusb_create_instance(btusb_t *btusb)
{
	struct usb_device	*usb_dev;
	int			err;

	DBFENTER;
	usb_dev = usb_alloc_dev(btusb->usb_dev, btusb->usb_dev->bus);
	if (!usb_dev) {
		DBPRT("Unable to allocate device\n");
		return -ENOMEM;
	}

	DBPRT("fake usb_dev: %p\n", usb_dev);

	//usb_connect(usb_dev);
	usb_dev->devnum = btusb->usb_dev->devnum;

	/* USB v1.1 5.5.3 */
	/* We read the first 8 bytes from the device descriptor to get to */
	/*  the bMaxPacketSize0 field. Then we set the maximum packet size */
	/*  for the control pipe, and retrieve the rest */
	usb_dev->epmaxpacketin [0] = 8;
	usb_dev->epmaxpacketout[0] = 8;

	err = usb_get_descriptor(usb_dev, USB_DT_DEVICE, 0, &usb_dev->descriptor, 8);
	if (err < 8) {
		if (err < 0)
			err("USB device not responding, giving up (error=%d)", err);
		else
			err("USB device descriptor short read (expected %i, got %i)", 8, err);
		goto fail;
	}
	usb_dev->epmaxpacketin [0] = usb_dev->descriptor.bMaxPacketSize0;
	usb_dev->epmaxpacketout[0] = usb_dev->descriptor.bMaxPacketSize0;

	err = usb_get_device_descriptor(usb_dev);
	if (err < (signed)sizeof(usb_dev->descriptor)) {
		if (err < 0)
			err("unable to get device descriptor (error=%d)", err);
		else
			err("USB device descriptor short read (expected %Zi, got %i)",
				sizeof(usb_dev->descriptor), err);
		goto fail;
	}

	err = usb_get_configuration(usb_dev);
	if (err < 0) {
		err("unable to get device %d configuration (error=%d)",
			usb_dev->devnum, err);
		goto fail;
	}

	/* we set the default configuration here */
	err = usb_set_configuration(usb_dev, usb_dev->config[0].bConfigurationValue);
	if (err) {
		err("failed to set device %d default configuration (error=%d)",
			usb_dev->devnum, err);
		goto fail;
	}

	dbg("new device strings: Mfr=%d, Product=%d, SerialNumber=%d",
		usb_dev->descriptor.iManufacturer,
		usb_dev->descriptor.iProduct,
		usb_dev->descriptor.iSerialNumber);

	btusb->usb_dev->children[0] = usb_dev;
	usb_driver_claim_interface(&btusb_driver, &usb_dev->actconfig->interface[0], btusb);
	DBFEXIT;
	return 0;
fail:
	//clear_bit(usb_dev->devnum, &usb_dev->bus->devmap.devicemap);
	//usb_dev->devnum = -1;	//FIXME
	usb_free_dev (usb_dev);
	DBPRT("Unable to init device\n");
	return -EIO;
}

void btusb_destroy_instance(struct usb_device *usb_dev, btusb_t *btusb)
{
	DBFENTER;
	/* master device */
	if (usb_dev->children[0]) {
		/* release fake interface */
		DBPRT("release fake usb_dev: %p\n", usb_dev->children[0]);
		usb_driver_release_interface(&btusb_driver, usb_dev->children[0]->actconfig->interface);
		usb_free_dev(usb_dev->children[0]);
		usb_dev->children[0] = NULL;
	}
	DBFEXIT;
}

#endif	/* BTUSB_CLAIM_OPEN */

static void *btusb_probe(struct usb_device *usb_dev, unsigned int ifnum, const struct usb_device_id *id)
{
	btusb_t		*btusb = NULL;
	int		btnum;
	int		i, ep, ep_found, vendor_id, device_id;
	btusb_req_t	*req = NULL;
	struct usb_interface_descriptor	*iface;

	if (ifnum)
		return NULL;

	DBFENTER;
	
	DBPRT("usb_dev: %p, ifnum: %d\n", usb_dev, ifnum);
		
	for (btnum = 0; btnum < NUM_OF_USB_BT_DEV; btnum++) {
		vendor_id = usb_bt_list[btnum].vendor_id;
		device_id = usb_bt_list[btnum].device_id;
		if( (vendor_id == usb_dev->descriptor.idVendor &&
		    device_id == usb_dev->descriptor.idProduct) || 
				(vendor_id == 0 && device_id == 0) )
			break;
	}
	if (btnum == NUM_OF_USB_BT_DEV)	{
		BTINFO("bluetooth device not supported\n");
		goto failed;
	}

#define EP(x) (iface->endpoint[x])

#ifdef CONFIG_AFFIX_DEBUG
	DBPRT("Number of Interfaces: %d\n", usb_dev->actconfig->bNumInterfaces);
	for (i = 0; i < usb_dev->actconfig->bNumInterfaces; i++) {
		int	s;
		DBPRT("   Interface: %d\n", i);
		DBPRT("   Number of altsettings: %d\n", usb_dev->actconfig->interface[i].num_altsetting);
		for (s = 0; s < usb_dev->actconfig->interface[i].num_altsetting; s++) {
			iface = &usb_dev->actconfig->interface[i].altsetting[s];
			DBPRT("      Altsetting: %d\n", s);
			DBPRT("      Number of endpoints: %d\n", iface->bNumEndpoints);
			for (ep = 0; ep < iface->bNumEndpoints; ep++) {
				DBPRT("        EndpointAddress: %#x, Attributes: %#x, Interval: %d\n",
				EP(ep).bEndpointAddress, EP(ep).bmAttributes, EP(ep).bInterval); 

			}
		}
	}
#endif
	/* allocating device object */
	btusb = kmalloc(sizeof(btusb_t), GFP_KERNEL);
	if (btusb == NULL) {
		BTERROR("failed to reserve memory for dev\n");
		goto failed;
	}
	memset(btusb, 0, sizeof(btusb_t));

	btusb->signature = 0xDEADBEAF;
	btusb->usb_dev = usb_dev;
	btusb->params = &usb_bt_list[btnum];

	/* initialize some data */
	btusb_init_req(&btusb->bulk_out_req, btusb);
	btusb_init_req(&btusb->bulk_in_req, btusb);
	btusb_init_req(&btusb->ctrl_out_req, btusb);
	btusb_init_req(&btusb->int_in_req, btusb);
	btusb_init_req(&btusb->isoc_out_req, btusb);
	btusb_init_req(&btusb->isoc_in_req, btusb);

	ep_found = 0;
	DBPRT("Checking interfaces...\n");
	for (i = 0; i < usb_dev->actconfig->bNumInterfaces; i++) {
		iface = &usb_dev->actconfig->interface[i].altsetting[0];
		for (ep = 0; ep < iface->bNumEndpoints; ep++) {
			switch (EP(ep).bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) {
				case USB_ENDPOINT_XFER_BULK:
					if ((EP(ep).bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN) {
						req = &btusb->bulk_in_req;
						req->pipe = usb_rcvbulkpipe(usb_dev, EP(ep).bEndpointAddress);
						ep_found |= 0x01;
					} else {
						req = &btusb->bulk_out_req;
						req->pipe = usb_sndbulkpipe(usb_dev, EP(ep).bEndpointAddress);
						ep_found |= 0x02;
					}
					break;
				case USB_ENDPOINT_XFER_INT:
					if ((EP(ep).bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN) {
						req = &btusb->int_in_req;
						req->pipe = usb_rcvintpipe(usb_dev, EP(ep).bEndpointAddress);
						btusb->int_in_req.interval = EP(ep).bInterval;
						ep_found |= 0x04;
					}
					break;
				case USB_ENDPOINT_XFER_ISOC:
					if ((EP(ep).bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN) {
						req = &btusb->isoc_in_req;
						req->pipe = usb_rcvisocpipe(usb_dev, EP(ep).bEndpointAddress);
						ep_found |= 0x10;
					} else {
						req = &btusb->isoc_out_req;
						req->pipe = usb_sndisocpipe(usb_dev, EP(ep).bEndpointAddress);
						ep_found |= 0x20;
					}
					break;
				default:
					req = NULL;
			}
		}
	}
	if ((ep_found & 0x0f) != 0x07) {
		BTERROR("invalid endpoint address / attribute\n");
		goto failed;
	}
	btusb->ctrl_out_req.pipe = usb_sndctrlpipe(usb_dev, 0);

	if ((ep_found & 0xF0) == 0x30) {
		DBPRT("Found ISOC interface\n");
		btusb->isoc = 1;
#ifndef BTUSB_CLAIM_OPEN
		usb_driver_claim_interface(&btusb_driver, &usb_dev->actconfig->interface[1], btusb);
		usb_set_interface(usb_dev, 1, 2);
#endif
		DBPRT("ISOC interface configured\n");
	}

#undef EP

	/* setting up hci inetrface */
	if ((btusb->hci = hcidev_alloc()) == NULL)
		goto failed;
	btusb->hci->priv = btusb;
	btusb->hci->open = btusb_net_open;
	btusb->hci->close = btusb_net_stop;
	btusb->hci->send = btusb_net_xmit_start;
	btusb->hci->ioctl = btusb_net_ioctl;
	btusb->hci->hdrlen = btusb->params->uart_leading_byte;
	btusb->hci->type = HCI_USB;
	SET_MODULE_OWNER(btusb->hci);

#ifdef	BTUSB_CLAIM_OPEN
	/* create another instance */
	if (btusb_create_instance(btusb) != 0) {
		BTERROR("register_netdev failed\n");
		goto failed2;
	}
#endif
	if (hcidev_register(btusb->hci, usb_dev) != 0) {
		BTERROR("register_netdev failed\n");
		goto failed3;
	}

	BTINFO("affix_usb bound to device\n");

	DBFEXIT;
#ifdef BTUSB_CLAIM_OPEN
	return NULL;
#else
	return btusb;
#endif
failed3:
#ifdef	BTUSB_CLAIM_OPEN
	btusb_destroy_instance(usb_dev, btusb);
failed2:
#endif
	kfree(btusb->hci);
failed:
	if (btusb) 
		kfree(btusb);
	return NULL;
}

static void btusb_disconnect(struct usb_device *usb_dev, void *ptr)
{
	btusb_t *btusb = (btusb_t *)ptr;

	DBFENTER;
	DBPRT("usb_dev: %p\n", usb_dev);
	if (btusb == NULL || btusb->signature != 0xDEADBEAF) {
		BTERROR("invalid btusb\n");
		return;
	}
#ifdef BTUSB_CLAIM_OPEN
	DBPRT("usb_dev: %p, fake: %p\n", btusb->usb_dev, btusb->usb_dev->children[0]);
	if (usb_dev == btusb->usb_dev)
		btusb_destroy_instance(usb_dev, btusb);
	else if (usb_dev == btusb->usb_dev->children[0]) {
		/* unloading */
		usb_driver_claim_interface(&btusb_driver, &btusb->usb_dev->actconfig->interface[0], btusb);
		return;
	}
#endif
	hcidev_unregister(btusb->hci);
	kfree(btusb);
	BTINFO("BTUSB driver disconnected from device\n");
	DBFEXIT;
}

/*----------------------------------------------------------------------------------------------
				Module init and shutdown functions
  ----------------------------------------------------------------------------------------------
*/

int __init init_btusb(void)
{
	int err;

	DBFENTER;

	printk("Affix USB Bluetooth driver loaded (affix_usb)\n");
	printk("Copyright (C) 2001, 2002 Nokia Corporation\n");
	printk("Written by Imre Deak <ext-imre.deak@nokia.com> and\n");
	printk("Dmitry Kasatkin <dmitry.kasatkin@nokia.com>\n");

	err = usb_register(&btusb_driver);
	if (err < 0) {
		BTERROR("usb_register failed for the USB bluetooth driver. Error number %d\n", err);
		DBFEXIT;
		return -1;
	}
	BTINFO("affix_usb registered\n");

	DBFEXIT;
	return 0;
}


void __exit exit_btusb(void)
{
	DBFENTER;
	usb_deregister(&btusb_driver);
	BTINFO("affix_usb deregistered\n");
	DBFEXIT;
}

/* ------------------------------------------------------------------------------------------ */


/*  If we are resident in kernel we will call module_init manually. */
#ifdef MODULE
module_init(init_btusb);
module_exit(exit_btusb);

MODULE_AUTHOR("Imre Deak/Dmitry Kasatkin");
MODULE_DESCRIPTION("Affix USB driver for Bluetooth devices");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 15)
MODULE_LICENSE("GPL");
#endif
#endif

EXPORT_NO_SYMBOLS;

